import pytest
from unittest.mock import patch

from geobox.tile3d import Tile3d
from geobox.user import User

def test_init(api, mock_tile3d_data):
    tile3d = Tile3d(api, mock_tile3d_data['uuid'], mock_tile3d_data)
    assert tile3d.name == mock_tile3d_data['name']
    assert tile3d.uuid == mock_tile3d_data['uuid']
    assert tile3d.data == mock_tile3d_data
    assert tile3d.endpoint == f'{Tile3d.BASE_ENDPOINT}{tile3d.uuid}/'


def test_get_3dtiles(api, mock_tile3d_data):
    api.get.return_value = [mock_tile3d_data, mock_tile3d_data]
    tile3ds = Tile3d.get_3dtiles(api)
    api.get.assert_called_once_with(f'{Tile3d.BASE_ENDPOINT}?f=json&return_count=False&skip=0&limit=10&shared=False')
    assert len(tile3ds) == 2
    assert type(tile3ds[0]) == Tile3d
    assert tile3ds[0].data == mock_tile3d_data


def test_get_3dtile(api, mock_tile3d_data):
    api.get.return_value = mock_tile3d_data
    tile3d = Tile3d.get_3dtile(api, uuid=mock_tile3d_data['uuid'])
    api.get.assert_called_once_with(f"{Tile3d.BASE_ENDPOINT}{mock_tile3d_data['uuid']}/?f=json")
    assert type(tile3d) == Tile3d
    assert tile3d.uuid == mock_tile3d_data['uuid']
    assert tile3d.data == mock_tile3d_data


def test_get_3dtile_by_name(api, mock_tile3d_data):
    api.get.return_value = [mock_tile3d_data]
    tile3d = Tile3d.get_3dtile_by_name(api, name=mock_tile3d_data['name'])
    api.get.assert_called_once_with(f"{Tile3d.BASE_ENDPOINT}?f=json&q=name+%3D+%27abasabad%27&return_count=False&skip=0&limit=10&shared=False")
    assert type(tile3d) == Tile3d
    assert tile3d.uuid == mock_tile3d_data['uuid']
    assert tile3d.data == mock_tile3d_data
    # not found
    tile3d = Tile3d.get_3dtile_by_name(api, name='not_found')
    assert tile3d == None


def test_update(api, mock_tile3d_data):
    tile3d = Tile3d(api, mock_tile3d_data['uuid'], mock_tile3d_data)
    updated_data = {**mock_tile3d_data,
                    'settings': {'settings': 'value'},
                    'thumbnail': 'thumbnail-url'}

    api.put.return_value = updated_data
    tile3d.update(settings={'settings': 'value'}, thumbnail='thumbnail-url')
    api.put.assert_called_once_with(tile3d.endpoint, {'settings': {'settings': 'value'}, 'thumbnail': 'thumbnail-url'})


def test_delete(api, mock_tile3d_data):
    tile3d = Tile3d(api, mock_tile3d_data['uuid'], mock_tile3d_data)
    endpoint = tile3d.endpoint
    tile3d.delete()
    api.delete.assert_called_once_with(endpoint)
    assert tile3d.uuid is None
    assert tile3d.endpoint is None


def test_thumbnail(api, mock_tile3d_data):
    tile3d = Tile3d(api, mock_tile3d_data['uuid'], mock_tile3d_data)
    thumbnail_url = tile3d.thumbnail
    assert thumbnail_url == f"{api.base_url}{tile3d.endpoint}thumbnail.png"


def test_get_item(api, mock_tile3d_data):
    tile3d = Tile3d(api, mock_tile3d_data['uuid'], mock_tile3d_data)
    api.get.return_value = {'data': 'test'}
    item = tile3d.get_item(path='test-path')
    api.get.assert_called_once_with(f"{tile3d.endpoint}test-path")
    assert item == {'data': 'test'}


def test_share(api, mock_tile3d_data, mock_user_data):
    """Test file sharing."""
    with patch.object(api, 'post'):
        tile3d = Tile3d(api, uuid=mock_tile3d_data['uuid'], data=mock_tile3d_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        tile3d.share(users=users)
        api.post.assert_called_once_with(
            f'{tile3d.endpoint}share/',
            {'user_ids': [1, 2]}, is_json=False
        )


def test_unshare(api, mock_tile3d_data, mock_user_data):
    """Test file unsharing."""
    with patch.object(api, 'post'):
        tile3d = Tile3d(api, uuid=mock_tile3d_data['uuid'], data=mock_tile3d_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        tile3d.unshare(users=users)
        
        # Verify the API call
        api.post.assert_called_once_with(
            f'{tile3d.endpoint}unshare/',
            {'user_ids': [1, 2]}, is_json=False
        )


def test_get_shared_users(api, mock_tile3d_data, mock_user_data):
    """Test getting shared users."""
    mock_response = [
        mock_user_data['1'],
        mock_user_data['2']
    ]
    
    with patch.object(api, 'get', return_value=mock_response):
        tile3d = Tile3d(api, uuid=mock_tile3d_data['uuid'], data=mock_tile3d_data)
        result = tile3d.get_shared_users(search='user', limit=2)

        api.get.assert_called_once_with(
            f'{tile3d.endpoint}shared-with-users/?search=user&skip=0&limit=2'
        )

        assert len(result) == 2
        assert result[0].first_name == 'test 1'
        assert result[0].last_name == 'test 1'
        assert result[1].first_name == 'test 2'
        assert result[1].last_name == 'test 2'


def test_get_tileset_json(api, mock_tile3d_data):
    mock_response = {
        'key': 'value'
    }
    with patch.object(api, 'get', return_value=mock_response):
        tile3d = Tile3d(api, uuid=mock_tile3d_data['uuid'], data=mock_tile3d_data)
        result = tile3d.get_tileset_json()
        
        assert result == mock_response

        api.get.assert_called_once_with('3dtiles/fa1e14c3-bb36-4541-a7c3-d0a7f70d8852/tileset.json')


def test_to_async(api, async_api, mock_tile3d_data):
    tile3d = Tile3d(api, uuid=mock_tile3d_data['uuid'], data=mock_tile3d_data)
    async_instance = tile3d.to_async(async_api)
    assert async_instance.api == async_api


from typing import List, Dict, Optional, TYPE_CHECKING, Union
from urllib.parse import urljoin

from .base import Base

if TYPE_CHECKING:
    from . import GeoboxClient
    from .aio import AsyncGeoboxClient
    from .aio.plan import AsyncPlan

class Plan(Base):

    BASE_ENDPOINT = 'plans/'

    def __init__(self,
                 api: 'GeoboxClient',
                 plan_id: int,
                 data: Optional[Dict] = {}):
        """
        Initialize a plan instance.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            plan_id (str): The id for the plan.
            data (Dict, optional): The data of the plan.
        """
        super().__init__(api, data=data)
        self.plan_id = plan_id
        self.endpoint = urljoin(self.BASE_ENDPOINT, str(self.id))


    @classmethod
    def get_plans(cls, api: 'GeoboxClient', **kwargs) -> Union[List['Plan'], int]:
        """
        Get list of plans with optional filtering and pagination.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.

        Keyword Args:
            q (str): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            search (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields (str): comma separated list of fields for searching.
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count (bool): Whether to return total count. default is False.
            skip (int): Number of items to skip. default is 0.
            limit (int): Number of items to return. default is 10.
            user_id (int): Specific user. privileges required.
            shared (bool): Whether to return shared plans. default is False.

        Returns:
            List[Plan] | int: A list of plan instances or the total number of plans.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.plan import Plan
            >>> client = GeoboxClient()
            >>> plans = Plan.get_plan(client, q="name LIKE '%My plan%'")
            or
            >>> plans = client.get_plan(q="name LIKE '%My plan%'")
        """
        params = {
           'f': 'json',
           'q': kwargs.get('q'),
           'search': kwargs.get('search'),
           'search_fields': kwargs.get('search_fields'),
           'order_by': kwargs.get('order_by'),
           'return_count': kwargs.get('return_count', False),
           'skip': kwargs.get('skip', 0),
           'limit': kwargs.get('limit', 10),
           'user_id': kwargs.get('user_id'),
           'shared': kwargs.get('shared', False)
        }
        return super()._get_list(api, cls.BASE_ENDPOINT, params, factory_func=lambda api, item: Plan(api, item['id'], item))


    @classmethod
    def create_plan(cls, 
                    api: 'GeoboxClient', 
                    name: str,
                    plan_color: str,
                    storage: int,
                    concurrent_tasks: int,
                    daily_api_calls: int,
                    monthly_api_calls: int,
                    daily_traffic: int,
                    monthly_traffic: int,
                    daily_process: int,
                    monthly_process: int,
                    number_of_days: int = None,
                    display_name: str = None,
                    description: str = None) -> 'Plan':
        """
        Create a new plan.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            name (str): The name of the plan.
            plan_color (str): hex value of the color. e.g. #000000.
            storage (int): storage value in bytes. must be greater that 1.
            concurrent_tasks (int): number of concurrent tasks. must be greater that 1.
            daily_api_calls (int): number of daily api calls. must be greater that 1.
            monthly_api_calls (int): number of monthly api calls. must be greater that 1.
            daily_traffic (int): number of daily traffic. must be greater that 1.
            monthly_traffic (int): number of monthly traffic. must be greater that 1.
            daily_process (int): number of daily processes. must be greater that 1.
            monthly_process (int): number of monthly processes. must be greater that 1.
            number_of_days (int, optional): number of days. must be greater that 1.
            display_name (str, optional): display name of the plan.
            description (str, optional): description of the plan.

        Returns:
            Plan: The newly created plan instance.

        Raises:
            ValidationError: If the plan data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.plan import Plan
            >>> client = GeoboxClient()
            >>> plan = Plan.create_plan(client, 
            ...                             name="new_plan",
            ...                             display_name=" New Plan",
            ...                             description="new plan description",
            ...                             plan_color="#000000",
            ...                             storage=10,
            ...                             concurrent_tasks=10,
            ...                             daily_api_calls=10,
            ...                             monthly_api_calls=10,
            ...                             daily_traffic=10,
            ...                             monthly_traffic=10,
            ...                             daily_process=10,
            ...                             monthly_process=10,
            ...                             number_of_days=10)
            or
            >>> plan = client.create_plan(name="new_plan",
            ...                             display_name=" New Plan",
            ...                             description="new plan description",
            ...                             plan_color="#000000",
            ...                             storage=10,
            ...                             concurrent_tasks=10,
            ...                             daily_api_calls=10,
            ...                             monthly_api_calls=10,
            ...                             daily_traffic=10,
            ...                             monthly_traffic=10,
            ...                             daily_process=10,
            ...                             monthly_process=10,
            ...                             number_of_days=10)
        """
        data = {
            "name": name,
            "display_name": display_name,
            "description": description,
            "plan_color": plan_color,
            "storage": storage,
            "concurrent_tasks": concurrent_tasks,
            "daily_api_calls": daily_api_calls,
            "monthly_api_calls": monthly_api_calls,
            "daily_traffic": daily_traffic,
            "monthly_traffic": monthly_traffic,
            "daily_process": daily_process,
            "monthly_process": monthly_process,
            "number_of_days": number_of_days
        }
        return super()._create(api, cls.BASE_ENDPOINT, data, factory_func=lambda api, item: Plan(api, item['id'], item))


    @classmethod
    def get_plan(cls, api: 'GeoboxClient', plan_id: int) -> 'Plan':
        """
        Get a plan by its id.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            plan_id (int): The id of the plan to get.

        Returns:
            Plan: The plan object

        Raises:
            NotFoundError: If the plan with the specified id is not found.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.plan import Plan
            >>> client = GeoboxClient()
            >>> plan = Plan.get_plan(client, plan_id=1)
            or
            >>> plan = client.get_plan(plan_id=1)
        """
        params = {
            'f': 'json'
        }
        return super()._get_detail(api, cls.BASE_ENDPOINT, plan_id, params, factory_func=lambda api, item: Plan(api, item['id'], item))
    

    @classmethod
    def get_plan_by_name(cls, api: 'GeoboxClient', name: str) -> Union['Plan', None]:
        """
        Get a plan by name

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            name (str): the name of the plan to get

        Returns:
            Plan | None: returns the plan if a plan matches the given name, else None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.plan import Plan
            >>> client = GeoboxClient()
            >>> plan = Plan.get_plan_by_name(client, name='test')
            or
            >>> plan = client.get_plan_by_name(name='test')
        """
        plans = cls.get_plans(api, q=f"name = '{name}'")
        if plans and plans[0].name == name:
            return plans[0]
        else:
            return None


    def update(self, **kwargs) -> Dict:
        """
        Update the plan

        Keyword Args:
            name (str): The name of the plan.
            plan_color (str): hex value of the color. e.g. #000000.
            storage (int): storage value in bytes. must be greater that 1.
            concurrent_tasks (int): number of concurrent tasks. must be greater that 1.
            daily_api_calls (int): number of daily api calls. must be greater that 1.
            monthly_api_calls (int): number of monthly api calls. must be greater that 1.
            daily_traffic (int): number of daily traffic. must be greater that 1.
            monthly_traffic (int): number of monthly traffic. must be greater that 1.
            daily_processes (int): number of daily processes. must be greater that 1.
            monthly_processes (int): number of monthly processes. must be greater that 1.
            number_of_days (int): number of days. must be greater that 1.
            display_name (str): display name of the plan.
            description (str): description of the plan.

        Returns:
            Dict: The updated plan data.

        Raises:
            ValidationError: If the plan data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.plan import Plan
            >>> client = GeoboxClient()
            >>> plan = Plan.get_plan(client, plan_id=1)
            >>> plan.update(display_name="New Display Name")
        """
        data = {
            "name": kwargs.get('name'),
            "display_name": kwargs.get('display_name'),
            "description": kwargs.get('description'),
            "plan_color": kwargs.get('plan_color'),
            "storage": kwargs.get('storage'),
            "concurrent_tasks": kwargs.get('concurrent_tasks'),
            "daily_api_calls": kwargs.get('daily_api_calls'),
            "monthly_api_calls": kwargs.get('monthly_api_calls'),
            "daily_traffic": kwargs.get('daily_traffic'),
            "monthly_traffic": kwargs.get('monthly_traffic'),
            "daily_process": kwargs.get('daily_process'),
            "monthly_process": kwargs.get('monthly_process'),
            "number_of_days": kwargs.get('number_of_days')
        }
        return super()._update(self.endpoint, data)


    def delete(self) -> None:
        """
        Delete the plan.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.plan import Plan
            >>> client = GeoboxClient()
            >>> plan = Plan.get_plan(client, plan_id=1)
            >>> plan.delete()
        """
        super()._delete(self.endpoint)
        self.plan_id = None


    def to_async(self, async_client: 'AsyncGeoboxClient') -> 'AsyncPlan':
        """
        Switch to async version of the plan instance to have access to the async methods

        Args:
            async_client (AsyncGeoboxClient): The async version of the GeoboxClient instance for making requests.

        Returns:
            AsyncPlan: the async instance of the plan.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.plan import Plan
            >>> client = GeoboxClient()
            >>> plan = Plan.get_plan(client, plan_id=1)
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     async_plan = plan.to_async(async_client)
        """
        from .aio.plan import AsyncPlan

        return AsyncPlan(api=async_client, plan_id=self.plan_id, data=self.data)


from typing import List, Dict, Optional, TYPE_CHECKING, Union
from urllib.parse import urljoin

from .base import Base
from .utils import clean_data
from .map import Map
from .vectorlayer import VectorLayer
from .view import VectorLayerView
from .file import File

if TYPE_CHECKING:
    from . import GeoboxClient 
    from .feature import Feature
    from .aio import AsyncGeoboxClient
    from .aio.attachment import AsyncAttachment


class Attachment(Base):

    BASE_ENDPOINT = 'attachments/'

    def __init__(self, 
                 api: 'GeoboxClient', 
                 attachment_id: str,
                 data: Optional[Dict] = {}):
        """
        Initialize an Attachment instance.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            attachment_id (str): The id for the attachment.
            data (Dict, optional): The data of the attachment.
        """
        super().__init__(api, data=data)
        self.attachment_id = attachment_id
        self.endpoint = f"{self.BASE_ENDPOINT}{str(self.attachment_id)}/"


    def __repr__(self) -> str:
        """
        Return a string representation of the attachment.

        Returns:
            str: The string representation of the attachment.
        """
        return f"Attachment(id={self.attachment_id}, name={self.name})"


    @property
    def file(self) -> 'File':
        """
        Attachment file property

        Returns:
            File: the file object
        """
        return File(self.api, self.data['file'].get('uuid'), self.data['file'])


    @classmethod
    def get_attachments(cls, api: 'GeoboxClient', resource: Union['Map', 'VectorLayer', 'VectorLayerView'], **kwargs) -> Union[List['Attachment'], int]:
        """
        Get list of attachments with optional filtering and pagination.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            resource (Map | VectorLayer | VectorLayerView): options are: Map, Vector, View objects

        Keyword Args:
            element_id (str): the id of the element with attachment.
            search (str): search term for keyword-based searching among all textual fields.
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            skip (int): Number of items to skip. default is 0.
            limit (int): Number of items to return. default is 10.
            return_count (bool): Whether to return total count. default is False.

        Returns:
            List[Attachment] | int: A list of attachments instances or the total number of attachments.

        Raises:
            TypeError: if the resource type is not supported

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.attachment import Attachment
            >>> client = GeoboxClient()
            >>> map = client.get_maps()[0]
            >>> attachments = Attachment.get_attachments(client, resource=map, q="name LIKE '%My attachment%'")
            or
            >>> attachments = client.get_attachments(resource=map, q="name LIKE '%My attachment%'")
        """
        if type(resource) == VectorLayer:
            resource_type = 'vector'

        elif type(resource) == VectorLayerView:
            resource_type = 'view'

        elif type(resource) == Map:
            resource_type = 'map'

        else:
            raise TypeError('resource must be a vectorlayer or view or map object')


        params = {
           'resource_type': resource_type,
           'resource_uuid': resource.uuid,
           'element_id': kwargs.get('element_id'),
           'search': kwargs.get('search'),
           'order_by': kwargs.get('order_by'),
           'skip': kwargs.get('skip', 0),
           'limit': kwargs.get('limit', 10),
           'return_count': kwargs.get('return_count')
        }
        return super()._get_list(api, cls.BASE_ENDPOINT, params, factory_func=lambda api, item: Attachment(api, item['id'], item))
    

    @classmethod
    def create_attachment(cls, 
                     api: 'GeoboxClient', 
                     name: str, 
                     loc_x: int,
                     loc_y: int,
                     resource: Union['Map', 'VectorLayer', 'VectorLayerView'],
                     file: 'File',
                     feature: 'Feature' = None,
                     display_name: str = None, 
                     description: str = None, ) -> 'Attachment':
        """
        Create a new Attachment.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            name (str): The name of the scene.
            loc_x (int): x parameter of the attachment location.
            loc_y (int): y parameter of the attachment location.
            resource (Map | VectorLayer | VectorLayerView): the resource object.
            file (File): the file object.
            feature (Feature, optional): the feature object.
            display_name (str, optional): The display name of the scene.
            description (str, optional): The description of the scene.

        Returns:
            Attachment: The newly created Attachment instance.

        Raises:
            ValidationError: If the Attachment data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.attachment import Attachment
            >>> client = GeoboxClient()
            >>> layer = client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>> feature = layer.get_feature(feature_id=1)
            >>> file = client.get_file(uuid="12345678-1234-5678-1234-567812345678")
            >>> attachment = Attachment.create_attachment(client, 
            ...                                             name="my_attachment", 
            ...                                             loc_x=30, 
            ...                                             loc_y=50, 
            ...                                             resource=layer, 
            ...                                             file=file, 
            ...                                             feature=feature, 
            ...                                             display_name="My Attachment", 
            ...                                             description="Attachment Description")
            or
            >>> attachment = client.create_attachment(name="my_attachment", 
            ...                                             loc_x=30, 
            ...                                             loc_y=50, 
            ...                                             resource=layer, 
            ...                                             file=file, 
            ...                                             feature=feature, 
            ...                                             display_name="My Attachment", 
            ...                                             description="Attachment Description")
        """
        if isinstance(resource, VectorLayer):
            resource_type = 'vector'

        if isinstance(resource, VectorLayerView):
            resource_type = 'view'

        if isinstance(resource, Map):
            resource_type = 'map'

        data = {
            "name": name,
            "display_name": display_name,
            "description": description,
            "loc_x": loc_x,
            "loc_y": loc_y,
            "resource_type": resource_type,
            "resource_uuid": resource.uuid,
            "element_id": feature.id if feature else None,
            "file_id": file.id
            }
        return super()._create(api, cls.BASE_ENDPOINT, data, factory_func=lambda api, item: Attachment(api, item['id'], item))


    @classmethod
    def update_attachment(cls, api: 'GeoboxClient', attachment_id: int, **kwargs) -> Dict:
        """
        Update the attachment.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            attachment_id (int): the attachment id.

        Keyword Args:
            name (str): The name of the attachment.
            display_name (str): The display name of the attachment.
            description (str): The description of the attachment.
            loc_x (int): x parameter of the attachment location.
            loc_y (int): y parameter of the attachment location.

        Returns:
            Dict: The updated attachment data.

        Raises:
            ValidationError: If the attachment data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.attachment import Attachment
            >>> client = GeoboxClient()
            >>> Attachment.update_attachment(client, attachment_id=1, display_name="New Display Name")
            or
            >>> client.update_attachment(attachment_id=1, display_name="New Display Name")
        """
        data = clean_data({
            "name": kwargs.get('name'),
            "display_name": kwargs.get('display_name'),   
            "description": kwargs.get('description'),
            "loc_x": kwargs.get('loc_x'),
            "loc_y": kwargs.get('loc_y')
        })
        endpoint = urljoin(cls.BASE_ENDPOINT, str(attachment_id))
        response = api.put(endpoint, data)
        return response
    

    def update(self, **kwargs) -> Dict:
        """
        Update the attachment.

        Keyword Args:
            name (str): The name of the attachment.
            display_name (str): The display name of the attachment.
            description (str): The description of the attachment.
            loc_x (int): x parameter of the attachment location.
            loc_y (int): y parameter of the attachment location.

        Returns:
            Dict: The updated attachment data.

        Raises:
            ValidationError: If the attachment data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.attachment import Attachment
            >>> client = GeoboxClient()
            >>> attachment = Attachment.get_attachments(client)[0]
            >>> attachment.update(display_name="New Display Name")
        """
        data = clean_data({
            "name": kwargs.get('name'),
            "display_name": kwargs.get('display_name'),   
            "description": kwargs.get('description'),
            "loc_x": kwargs.get('loc_x'),
            "loc_y": kwargs.get('loc_y')
        })
        response = self.api.put(self.endpoint, data)
        self._update_properties(response)
        return response
    

    def delete(self) -> None:
        """
        Delete the scene.

        Returns:
            None

        Raises:
            ApiRequestError: If the API request fails.
            ValidationError: If the scene data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.attachment import Attachment
            >>> client = GeoboxClient()
            >>> attachment = Attachment.get_attachments(client)[0]
            >>> attachment.delete()
        """
        super()._delete(self.endpoint)
        self.attachment_id = None


    @property
    def thumbnail(self) -> str:
        """
        Get the thumbnail URL of the attachment.

        Returns:
            str: The thumbnail of the scene.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.attachment import Attachment
            >>> client = GeoboxClient()
            >>> attachment = Attachment.get_attachment(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> attachment.thumbnail
        """
        return super()._thumbnail(format='')
    

    def to_async(self, async_client: 'AsyncGeoboxClient') -> 'AsyncAttachment':
        """
        Switch to async version of the attachment instance to have access to the async methods

        Args:
            async_client (AsyncGeoboxClient): The async version of the GeoboxClient instance for making requests.

        Returns:
            AsyncAttachment: the async instance of the attachment.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.attachment import Attachment
            >>> client = GeoboxClient()
            >>> attachment = Attachment.get_attachment(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     async_attachment = attachment.to_async(async_client)
        """
        from .aio.attachment import AsyncAttachment

        return AsyncAttachment(api=async_client, attachment_id=self.attachment_id, data=self.data)


import pytest
from urllib.parse import urlencode
from urllib.parse import urljoin
from unittest.mock import patch

from geobox.enums import FeatureType
from geobox.vectorlayer import VectorLayer, LayerType, InputGeomType, FileOutputFormat
from geobox.version import VectorLayerVersion
from geobox.field import Field, FieldType
from geobox.view import VectorLayerView
from geobox.user import User
from geobox.task import Task
from geobox.file import File
from geobox.attachment import Attachment
from geobox.exception import NotFoundError
from geobox.utils import clean_data


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_init(api, mock_vector_data, layer_type):
    """Test VectorLayer initialization."""
    layer = VectorLayer(api, 
                            uuid=mock_vector_data['uuid'], 
                            data=mock_vector_data,
                            layer_type=layer_type)
    assert layer.name == mock_vector_data["name"]
    assert layer.layer_type == layer_type
    assert layer.uuid == mock_vector_data["uuid"]
    assert layer.data == mock_vector_data
    assert layer.endpoint == f'{VectorLayer.BASE_ENDPOINT}{mock_vector_data["uuid"]}/'


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_repr(api, mock_vector_data, layer_type):
    """Test the repr method of Layer object for different geometry types"""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    assert repr(layer) == f"VectorLayer(uuid={layer.uuid}, name={mock_vector_data['name']}, layer_type={layer_type})"


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_get_vectors(api, mock_vector_data, layer_type):
    """Test getting a list of layers."""
    mock_vector_data['layer_type'] = layer_type.value
    api.get.return_value = [mock_vector_data, mock_vector_data]

    layers = VectorLayer.get_vectors(api)

    assert len(layers) == 2
    assert isinstance(layers[0], VectorLayer)
    assert layers[0].data == mock_vector_data
    assert layers[0].layer_type == layer_type
    assert layers[0].uuid == mock_vector_data['uuid']
    assert layers[0].endpoint == f"{VectorLayer.BASE_ENDPOINT}{mock_vector_data['uuid']}/"
    api.get.assert_called_once_with(f'{VectorLayer.BASE_ENDPOINT}?f=json&include_settings=False&temporary=False&return_count=False&skip=0&limit=10&shared=False')
    # return count
    api.reset_mock()
    api.get.return_value = 1
    count = VectorLayer.get_vectors(api, return_count=True)
    assert type(count) == int
    assert count == 1
    api.get.assert_called_once_with(f'{VectorLayer.BASE_ENDPOINT}?f=json&include_settings=False&temporary=False&return_count=True&skip=0&limit=10&shared=False')


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_get_vector(api, mock_vector_data, layer_type):
    """Test getting a specific layer."""
    mock_vector_data["layer_type"] = layer_type.value
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.get.return_value = mock_vector_data

    layer = VectorLayer.get_vector(api, mock_vector_data["uuid"])

    assert layer.name == mock_vector_data["name"]
    assert layer.layer_type.value == layer_type.value
    assert layer.uuid == mock_vector_data["uuid"]
    assert layer.data == mock_vector_data
    assert layer.endpoint == f'{VectorLayer.BASE_ENDPOINT}{mock_vector_data["uuid"]}/'
    api.get.assert_called_once_with(f'{layer.endpoint}?f=json')


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_get_vector_by_name(api, mock_vector_data, layer_type):
    """Test getting a specific layer."""
    mock_vector_data["layer_type"] = layer_type.value
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.get.return_value = [mock_vector_data]

    layer = VectorLayer.get_vector_by_name(api, name=mock_vector_data["name"])

    assert layer.name == mock_vector_data["name"]
    assert layer.layer_type.value == layer_type.value
    assert layer.uuid == mock_vector_data["uuid"]
    assert layer.data == mock_vector_data
    assert layer.endpoint == f'{VectorLayer.BASE_ENDPOINT}{mock_vector_data["uuid"]}/'
    api.get.assert_called_once_with(f'{VectorLayer.BASE_ENDPOINT}?f=json&include_settings=False&temporary=False&q=name+%3D+%27tehran_water%27&return_count=False&skip=0&limit=10&shared=False')
    # not found
    layer = VectorLayer.get_vector_by_name(api, name='not_found')
    assert layer == None


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_get_vectors_by_ids(api, mock_vector_data, layer_type):
    """Test the get_mosaics_by_ids method."""
    mock_response = [{**mock_vector_data, **{'id': 1}}, {**mock_vector_data, **{'id': 2}}]
    api.get.return_value = mock_response
    
    layers = VectorLayer.get_vectors_by_ids(api, [1, 2])
    
    api.get.assert_called_once_with(f'{VectorLayer.BASE_ENDPOINT}get-layers/?ids=%5B1%2C+2%5D&include_settings=False')
    assert len(layers) == 2
    assert all(isinstance(m, VectorLayer) for m in layers)
    assert layers[0].data == {**mock_vector_data, **{'id': 1}}
    assert layers[1].data == {**mock_vector_data, **{'id': 2}}



@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_create_vector(api, mock_vector_data, layer_type):
    """Test creating a new layer."""
    api.post.return_value = mock_vector_data

    layer = VectorLayer.create_vector(
        api,
        name=mock_vector_data["name"],
        layer_type=layer_type,
        display_name=mock_vector_data["display_name"]
    )

    expected_data = {
        "name": mock_vector_data["name"],
        "layer_type": layer_type.value,
        "has_z": mock_vector_data["has_z"],
        "temporary": False
    }

    api.post.assert_called_once_with('vectorLayers/', expected_data)
    assert layer.name == mock_vector_data["name"]
    assert layer.layer_type == layer_type
    assert layer.uuid == mock_vector_data["uuid"]
    assert layer.data == mock_vector_data
    assert layer.endpoint == f'{VectorLayer.BASE_ENDPOINT}{mock_vector_data["uuid"]}/'


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_update( api, mock_vector_data, layer_type):
    """Test updating layer properties."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    new_name = "updated_layer"
    new_display_name = "Updated Layer"
    
    # Response data - what we get back from the API
    api_response = {
        **mock_vector_data,
        'name': new_name,
        'display_name': new_display_name
    }
    api.put.return_value = api_response

    response = layer.update(
        name=new_name,
        display_name=new_display_name
    )

    # Only include the fields we're updating
    expected_data = {
        "name": new_name,
        "display_name": new_display_name
    }

    api.put.assert_called_once_with(layer.endpoint, expected_data)
    assert response == api_response
    assert layer.data == api_response


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_delete(api, mock_vector_data, layer_type):
    """Test deleting a layer."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    endpoint = layer.endpoint
    layer.delete()
    api.delete.assert_called_once_with(endpoint)
    assert layer.uuid is None
    assert layer.layer_type == layer_type
    assert layer.endpoint is None


def test_make_permanent(api, mock_vector_data):
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=LayerType.Point)
    layer.make_permanent()
    api.post.assert_called_once_with(f'{layer.endpoint}makePermanent/', is_json=False)


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_share(api, mock_vector_data, layer_type, mock_user_data):
    """Test file sharing."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    users = [
        User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
        User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
    ]
    layer.share(users=users)
    api.post.assert_called_once_with(
        f'{layer.endpoint}share/',
        {'user_ids': [1, 2]}, is_json=False
    )


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_unshare(api, mock_vector_data, layer_type, mock_user_data):
    """Test file unsharing."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    users = [
        User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
        User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
    ]
    layer.unshare(users=users)
    api.post.assert_called_once_with(
        f'{layer.endpoint}unshare/',
        {'user_ids': [1, 2]}, is_json=False
    )


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_get_shared_users(api, mock_vector_data, layer_type, mock_user_data):
    """Test getting shared users."""
    mock_response = [
        mock_user_data['1'],
        mock_user_data['2']
    ]
    api.get.return_value = mock_response
    
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    result = layer.get_shared_users(search='user', limit=2)

    api.get.assert_called_once_with(
        f'{layer.endpoint}shared-with-users/?search=user&skip=0&limit=2'
    )

    assert len(result) == 2
    assert result[0].first_name == 'test 1'
    assert result[0].last_name == 'test 1'
    assert result[1].first_name == 'test 2'
    assert result[1].last_name == 'test 2'


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_create_version(api, mock_vector_data, layer_type, mock_version_data):
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.post.return_value = mock_version_data
    version = layer.create_version(name=mock_version_data['name'], display_name=mock_version_data['display_name'], description=mock_version_data['description'])
    assert isinstance(version, VectorLayerVersion)
    assert version.uuid == mock_version_data['uuid']
    assert version.data == mock_version_data
    api.post.assert_called_once_with(f'{layer.endpoint}versions', {'name': 'design_version', 'display_name': 'Design Version', 'description': 'This layer represents design version.'})


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_get_versions(api, mock_vector_data, layer_type, mock_version_data):
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.get.return_value = [mock_version_data]
    versions = layer.get_versions()
    assert len(versions) == 1
    assert versions[0].uuid == mock_version_data['uuid']
    assert versions[0].data == mock_version_data
    api.get.assert_called_once_with(f'vectorLayerVersions/?layer_id={layer.id}&f=json&return_count=False&skip=0&limit=10&shared=False')


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_wfs_access_token(api, mock_vector_data, layer_type):
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    wfs_url = layer.wfs
    assert wfs_url == f'{api.base_url}{layer.endpoint}wfs/'


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_wfs_apikey(api, mock_vector_data, layer_type):
    api.access_token = None
    api.apikey = '123456789'
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    wfs_url = layer.wfs
    assert wfs_url == f'{api.base_url}{layer.endpoint}apikey:{api.apikey}/wfs/'


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_get_fields(api, mock_vector_data, layer_type, mock_field_data):
    """Test getting the fields of a layer."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.get.reset_mock()
    api.get.return_value = [mock_field_data, mock_field_data]
    fields = layer.get_fields()
    assert len(fields) == 2
    assert type(fields[0]) == Field
    assert fields[0].data == mock_field_data
    api.get.assert_called_once_with(f'{layer.endpoint}fields/')


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_get_field(api, mock_vector_data, layer_type, mock_field_data):
    """Test getting the fields of a layer."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.get.return_value = [mock_field_data, mock_field_data]
    field = layer.get_field(field_id=mock_field_data['id'])
    assert type(field) == Field
    assert field.data == mock_field_data
    api.get.assert_called_once_with(f'{layer.endpoint}fields/')


    assert type(field) == Field
    assert field.data == mock_field_data
    api.get.assert_called_once_with(f'{layer.endpoint}fields/')

    # error
    api.get.return_value = []
    with pytest.raises(NotFoundError):
        layer.get_field(field_id=mock_field_data['id'])



@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_get_field_by_name(api, mock_vector_data, layer_type, mock_field_data):
    """Test getting the fields of a layer."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.get.return_value = [mock_field_data, mock_field_data]
    field = layer.get_field_by_name(name=mock_field_data['name'])

    assert type(field) == Field
    assert field.data == mock_field_data
    api.get.assert_called_once_with(f'{layer.endpoint}fields/')

    # error
    api.get.return_value = []
    with pytest.raises(NotFoundError):
        layer.get_field_by_name(name=mock_field_data['name'])


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_add_field(api, mock_vector_data, layer_type, mock_field_data):
    """Test adding a field to a layer."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.post.return_value = mock_field_data
    del mock_field_data['created_at']
    field = layer.add_field(name=mock_field_data['name'], data_type=FieldType.String, data=mock_field_data)
    assert isinstance(field, Field)
    assert field.name == mock_field_data['name']
    assert field.data_type == FieldType.String
    assert field.layer == layer
    assert field.endpoint == f'{layer.endpoint}fields/{field.id}/'
    assert field.data == mock_field_data
    expected_data = {'id': mock_field_data['id'], 
                        'name': mock_field_data['name'], 
                        'datatype': mock_field_data['datatype'], 
                        'hyperlink': mock_field_data['hyperlink']}
    api.post.assert_called_once_with(f'{layer.endpoint}fields/', expected_data)


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_calculate_field(api, mock_vector_data, layer_type, mock_success_task_data):
    """Test calculating a field."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.get.return_value = mock_success_task_data
    task = layer.calculate_field(target_field='test_field', expression='name + "test"', q='name like "test"', bbox=[10, 20, 30, 40], bbox_srid=3857, feature_ids=[1, 2, 3])
    expected_data = {'target_field': 'test_field', 'expression': 'name + "test"', 'q': 'name like "test"', 'bbox': [10, 20, 30, 40], 'bbox_srid': 3857, 'feature_ids': [1, 2, 3], 'run_async': True}
    api.post.assert_called_once_with(urljoin(layer.endpoint, 'calculateField/'), expected_data, is_json=False)
    assert isinstance(task, Task)


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_calculate_field_sync(api, mock_vector_data, layer_type):
    """Test calculating a field."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.post.return_value = {'key': 'value'}
    response = layer.calculate_field(target_field='test_field', expression='name + "test"', q='name like "test"', bbox=[10, 20, 30, 40], bbox_srid=3857, feature_ids=[1, 2, 3], run_async=False)
    expected_data = {'target_field': 'test_field', 'expression': 'name + "test"', 'q': 'name like "test"', 'bbox': [10, 20, 30, 40], 'bbox_srid': 3857, 'feature_ids': [1, 2, 3], 'run_async': False}
    api.post.assert_called_once_with(urljoin(layer.endpoint, 'calculateField/'), expected_data, is_json=False)
    assert isinstance(response, dict)
    assert response['key'] == 'value'


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_get_features(api, mock_vector_data, layer_type, mock_feature_data):
    """Test getting a specific feature."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    layer.create_feature(geojson=mock_feature_data[FeatureType.Polygon.value])
    geojson = {
        'type': 'FeatureCollection',
        'features': [mock_feature_data[FeatureType.Polygon.value]],
        'crs': {
            'properties': {
                'name': 'EPSG:3857'
            }
        }
    }
    api.get.return_value = geojson
    params = clean_data({
        'f': 'json',
        'quant_factor': 1000000,
        'skip': 0,
        'limit': 100,
        'user_id': 123,
        'skip_geometry': False,
        'return_count': False,
        'feature_ids': [mock_feature_data[FeatureType.Polygon.value]['id']],
        'select_fields': "[ALL]",
        'out_srid': 3857,
        'order_by': "name",
        'q': "name like 'test'",
        'bbox_srid': 3857
    })
    features = layer.get_features(**params)
    query_string = urlencode(params)
    expected_endpoint = f'vectorLayers/{mock_vector_data["uuid"]}/features/?{query_string}'
    api.get.assert_called_once_with(expected_endpoint)
    assert len(features) == 1
    assert features[0].id == mock_feature_data[FeatureType.Polygon.value]['id']
    assert features[0].layer == layer
    assert features[0].data == mock_feature_data[FeatureType.Polygon.value]
    assert features[0].endpoint == f'vectorLayers/{mock_vector_data["uuid"]}/features/{mock_feature_data[FeatureType.Polygon.value]["id"]}/'
    # get geojson
    result = layer.get_features(geojson=True)
    assert type(result) == dict
    assert result == geojson
    # return count
    api.get.return_value = {'count': 1}
    count = layer.get_features(return_count=True)
    assert type(count) == int
    assert count == 1


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_get_feature(api, mock_vector_data, layer_type, mock_feature_data):
    """Test getting a specific feature."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    layer.create_feature(geojson=mock_feature_data[FeatureType.Polygon.value])
    api.get.return_value = mock_feature_data[FeatureType.Polygon.value]
    feature = layer.get_feature(feature_id=int(mock_feature_data[FeatureType.Polygon.value]['id']), out_srid=4326)
    
    expected_endpoint = f'vectorLayers/{mock_vector_data["uuid"]}/features/{int(mock_feature_data[FeatureType.Polygon.value]["id"])}'
    api.get.assert_called_once_with(expected_endpoint)
    
    assert feature is not None
    assert feature.id == mock_feature_data[FeatureType.Polygon.value]['id']
    assert feature.layer == layer
    assert feature.srid == 4326
    assert feature.data == mock_feature_data[FeatureType.Polygon.value]
    assert feature.endpoint == f'vectorLayers/{mock_vector_data["uuid"]}/features/{mock_feature_data[FeatureType.Polygon.value]["id"]}/'


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_create_feature(api, mock_vector_data, layer_type, mock_feature_data):
    """Test creating a new feature."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.post.return_value = mock_feature_data[FeatureType.Polygon.value]
    feature = layer.create_feature(geojson=mock_feature_data[FeatureType.Polygon.value])
    assert feature is not None
    assert feature.id == mock_feature_data[FeatureType.Polygon.value]['id']
    assert feature.layer == layer
    assert feature.data == mock_feature_data[FeatureType.Polygon.value]
    assert feature.endpoint == f'vectorLayers/{mock_vector_data["uuid"]}/features/{mock_feature_data[FeatureType.Polygon.value]["id"]}/'


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_delete_features(api, mock_vector_data, layer_type, mock_feature_data, mock_success_task_data):
    """Test deleting a specific feature."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    params = {
        'q': "name like 'test'",
        'bbox': [10, 20, 30, 40],
        'bbox_srid': 3857,
        'feature_ids': [mock_feature_data[FeatureType.Polygon.value]['id']],
        'run_async': True,
        'user_id': 123
    }
    api.get.return_value = mock_success_task_data
    task = layer.delete_features(**params)
    assert isinstance(task, Task)
    expected_endpoint = f'vectorLayers/{mock_vector_data["uuid"]}/deleteFeatures/'
    api.post.assert_called_once_with(expected_endpoint, params, is_json=False)


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_delete_features_sync(api, mock_vector_data, layer_type, mock_feature_data):
    """Test deleting a specific feature."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.post.return_value = {'key': 'value'}
    params = {
        'q': "name like 'test'",
        'bbox': [10, 20, 30, 40],
        'bbox_srid': 3857,
        'feature_ids': [mock_feature_data[FeatureType.Polygon.value]['id']],
        'run_async': False,
        'user_id': 123
    }
    response = layer.delete_features(**params)
    assert isinstance(response, dict)
    assert response['key'] == 'value'
    expected_endpoint = f'vectorLayers/{mock_vector_data["uuid"]}/deleteFeatures/'
    api.post.assert_called_once_with(expected_endpoint, params, is_json=False)


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_import_features(api, mock_vector_data, layer_type, mock_success_task_data, mock_file_data):
    """Test importing features."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    file = File(api, mock_file_data['uuid'], mock_file_data)
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    task = layer.import_features(file=file, input_geom_type=InputGeomType.POINT, input_layer_name='test-layer', input_dataset='test-dataset', user_id=123, input_srid=3857, file_encoding='utf-8', replace_domain_codes_by_values=True, report_errors=True)
    expected_data = {
        'file_uuid': mock_file_data['uuid'],
        'input_layer': 'test-layer',
        'input_geom_type': 'POINT',
        'replace_domain_codes_by_values': True,
        'input_dataset': 'test-dataset',
        'user_id': 123,
        'input_srid': 3857,
        'file_encoding': 'utf-8',
        'report_errors': True
    }
    api.post.assert_called_once_with(f"{layer.endpoint}import/", expected_data, is_json=False)
    assert isinstance(task, Task)


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_export_features(api, mock_vector_data, layer_type, mock_success_task_data):
    """Test exporting features."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.get.return_value = mock_success_task_data
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    task = layer.export_features(out_filename='test-file-name', out_format=FileOutputFormat.Shapefile, replace_domain_codes_by_values=True, run_async=True, bbox=[10, 20, 30, 40], out_srid=3857, zipped=True, feature_ids=[1, 2, 3], bbox_srid=3857, q='name like "test"', fields=['name', 'description'])
    expected_data = {
        'replace_domain_codes_by_values': True,
        'out_format': 'Shapefile',
        'run_async': True,
        'bbox': [10, 20, 30, 40],
        'out_srid': 3857,
        'zipped': True,
        'feature_ids': [1, 2, 3],
        'bbox_srid': 3857,
        'q': 'name like "test"',
        'fields': ['name', 'description'],
        'out_filename': 'test-file-name'
    }
    api.post.assert_called_once_with(f"{layer.endpoint}export/", expected_data, is_json=False)
    assert isinstance(task, Task)


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_export_features_sync(api, mock_vector_data, layer_type):
    """Test exporting features."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.post.return_value = {'key': 'value'}
    response = layer.export_features(out_filename='test-file-name', out_format=FileOutputFormat.Shapefile, replace_domain_codes_by_values=True, run_async=False, bbox=[10, 20, 30, 40], out_srid=3857, zipped=True, feature_ids=[1, 2, 3], bbox_srid=3857, q='name like "test"', fields=['name', 'description'])
    expected_data = {
        'replace_domain_codes_by_values': True,
        'out_format': 'Shapefile',
        'run_async': False,
        'bbox': [10, 20, 30, 40],
        'out_srid': 3857,
        'zipped': True,
        'feature_ids': [1, 2, 3],
        'bbox_srid': 3857,
        'q': 'name like "test"',
        'fields': ['name', 'description'],
        'out_filename': 'test-file-name'
    }
    api.post.assert_called_once_with(f"{layer.endpoint}export/", expected_data, is_json=False)
    assert isinstance(response, dict)
    assert response['key'] == 'value'


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_create_view(api, mock_vector_data, layer_type, mock_view_data):
    """Test creating a new view."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.post.return_value = mock_view_data
    mock_view_data['vector_layer'] = mock_vector_data
    expected_data = {
        'name': mock_view_data['name'],
        'display_name': mock_view_data['display_name'],
        'description': mock_view_data['description'],
        'view_filter': mock_view_data['view_filter'], 
        'view_extent': mock_view_data['extent'], 
        'view_cols': mock_view_data['view_cols']
    }

    view = layer.create_view(**expected_data)

    api.post.assert_called_once_with(urljoin(layer.endpoint, 'views/'), expected_data)
    
    # Verify the view object
    assert isinstance(view, VectorLayerView)
    assert view.name == mock_view_data['name']
    assert view.description == mock_view_data['description']
    assert view.layer_type == layer_type
    assert view.uuid == mock_view_data['uuid']
    assert view.data == mock_view_data
    assert isinstance(view.vector_layer, VectorLayer)


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_get_tile_pbf_url(api, mock_vector_data, layer_type):
    """Test getting a vector tile."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    url = layer.get_tile_pbf_url(x=1, y=2, z=3)
    assert url == f'{layer.api.base_url}{layer.endpoint}tiles/3/1/2.pbf'
    # apikey
    api.access_token = ''
    api.apikey = 'apikey_1234'
    url = layer.get_tile_pbf_url(x=1, y=2, z=3)
    assert url == f'{layer.api.base_url}{layer.endpoint}tiles/3/1/2.pbf?apikey=apikey_1234'


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_get_tile_json(api, mock_vector_data, layer_type):
    """Test getting a vector tile JSON."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.get.return_value = {'tile': 'value'}
    result = layer.get_tile_json()
    api.get.assert_called_once_with(f'{layer.endpoint}tilejson.json')
    assert result == {'tile': 'value'}


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_settings(api, mock_vector_data, layer_type):
    """Test getting the settings of a layer."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.get.return_value = {'settings': 'value'}
    settings = layer.settings
    api.get.assert_called_once_with(f'{layer.endpoint}settings/?f=json')
    assert settings == {'settings': 'value'}


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_update_settings(api, mock_vector_data, layer_type):
    layer1 = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    layer2 = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    settings = layer2.settings
    api.put.return_value = settings
    layer1.update_settings(layer2.settings)
    api.put.assert_called_once_with(f'{layer1.endpoint}settings/', settings)


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_set_settings(api, mock_vector_data, layer_type):
    """Test setting the settings of a layer."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.get.return_value = {
        "general_settings": {
            "title_field": None,
            "domain_display_type": "Value",
            "allow_export": True
        },
        "edit_settings": {
            "editable": True,
            "edit_geometry": True,
            "editable_attributes": "[ALL]",
            "allow_insert": True,
            "allow_delete": True
        },
        "tile_settings": {
            "min_zoom": 0,
            "max_zoom": 24,
            "max_features": 65536,
            "filter_features": True,
            "fields": [
            "id"
            ],
            "use_cache": True,
            "cache_until_zoom": 17
        }
        }
    settings = {
        'title_field': 'name',
        'domain_display_type': ' value',
        'allow_export': True,
        'editable': True,
        'edit_geometry': True,
        'editable_attributes': ['name', 'description'],
        'allow_insert': True,
        'allow_delete': True,
        'min_zoom': 0,
        'max_zoom': 24,
        'max_features': 10000,
        'filter_features': True,
        'fields': ['name', 'description'],
        'use_cache': True,
        'cache_until_zoom': 17
    }

    layer.set_settings(**settings)

    expected_data = {
        'general_settings': {
            'title_field': settings['title_field'],
            'domain_display_type': settings['domain_display_type'],
            'allow_export': settings['allow_export']
        },
        'edit_settings': {
            'editable': settings['editable'],
            'edit_geometry': settings['edit_geometry'],
            'editable_attributes': settings['editable_attributes'],
            'allow_insert': settings['allow_insert'],
            'allow_delete': settings['allow_delete']
        },
        'tile_settings': {
            'min_zoom': settings['min_zoom'],
            'max_zoom': settings['max_zoom'],
            'max_features': settings['max_features'],
            'filter_features': settings['filter_features'],
            'fields': settings['fields'],
            'use_cache': settings['use_cache'],
            'cache_until_zoom': settings['cache_until_zoom']
        }
    }
    api.put.assert_called_once_with(f'{layer.endpoint}settings/', expected_data)


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_seed_cache(api, mock_vector_data, layer_type, mock_success_task_data):
    """Test caching a seed."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.post.return_value = [{'task_id': mock_success_task_data['id']}]
    api.get_task.return_value = Task(api, mock_success_task_data['uuid'], mock_success_task_data)
    task = layer.seed_cache(from_zoom=0, to_zoom=10, ignore_cache=False, workers=1)[0]
    expected_data = {'from_zoom': 0, 'to_zoom': 10, 'ignore_cache': False, 'workers': 1}
    api.post.assert_called_once_with(f'{layer.endpoint}cache/seed/', expected_data)
    assert isinstance(task, Task)
    # error
    with pytest.raises(ValueError, match="workers must be in \\[1, 2, 4, 8, 12, 16, 20, 24\\]"):
        layer.seed_cache(workers=5)

    api.post.return_value = {'ivalid_response': 'value', 'invalid_response': 'value'}
    with pytest.raises(ValueError, match="Failed to seed cache"):
        layer.seed_cache()


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_clear_cache(api, mock_vector_data, layer_type):
    """Test clearing the cache."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    layer.clear_cache()
    api.post.assert_called_once_with(f'{layer.endpoint}cache/clear/')


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_cache_size(api, mock_vector_data, layer_type):
    """Test getting the size of the cache."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    api.post.return_value = 1024
    size = layer.cache_size
    api.post.assert_called_once_with(f'{layer.endpoint}cache/size/')
    assert size == 1024


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_update_stats(api, mock_vector_data, layer_type):
    """Test updating the stats."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    layer.update_stats()
    api.post.assert_called_once_with(f'{layer.endpoint}updateStats/')


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_prune_edited_areas(api, mock_vector_data, layer_type):
    """Test pruning the edited areas."""
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    layer.prune_edited_areas()
    api.post.assert_called_once_with(f'{layer.endpoint}prune/')


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_get_attachments(api, mock_vector_data, mock_attachment_data, layer_type):
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)

    api.get.return_value = [mock_attachment_data, mock_attachment_data, mock_attachment_data]

    attachments = layer.get_attachments()
    assert len(attachments) == 3
    assert type(attachments[0]) == Attachment


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_create_attachment(api, mock_vector_data, mock_attachment_data, mock_file_data, layer_type):
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)

    file = File(api, uuid=mock_file_data['uuid'], data=mock_file_data)

    api.post.return_value = mock_attachment_data
    attachment = layer.create_attachment(name='test', loc_x=10, loc_y=10, file=file)
    assert type(attachment) == Attachment
    assert attachment.data == mock_attachment_data


@pytest.mark.parametrize("layer_type", [type for type in LayerType])
def test_to_async(api, async_api, mock_vector_data, layer_type):
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], data=mock_vector_data, layer_type=layer_type)
    async_instance = layer.to_async(async_api)
    assert async_instance.api == async_api



from urllib.parse import urljoin
from typing import List, Optional, TYPE_CHECKING, Dict, Union

from .base import AsyncBase
from .field import AsyncField, FieldType
from .feature import AsyncFeature
from .task import AsyncTask
from .version import AsyncVectorLayerVersion
from ..utils import clean_data
from ..exception import NotFoundError
from ..enums import LayerType, InputGeomType, FileOutputFormat

if TYPE_CHECKING:
    from .api import AsyncGeoboxClient
    from .view import AsyncVectorLayerView
    from .user import AsyncUser
    from .file import AsyncFile
    from .attachment import AsyncAttachment
    from ..api import GeoboxClient
    from ..vectorlayer import VectorLayer


class AsyncVectorLayer(AsyncBase):
    """
    A class representing a vector layer in Geobox.
    
    This class provides functionality to create, manage, and manipulate vector layers.
    It supports various operations including CRUD operations on layers, features, and fields,
    as well as advanced operations like importing/exporting features and calculating field values.
    """
    BASE_ENDPOINT = 'vectorLayers/'

    def __init__(self, 
        api: 'AsyncGeoboxClient', 
        uuid: str, 
        layer_type: LayerType, 
        data: Optional[Dict] = {}):
        """
        Initialize a VectorLayer instance.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            uuid (str): The unique identifier for the layer.
            layer_type (LayerType): The type of geometry stored in the layer.
            data (Dict, optional): Additional layer metadata and configuration.
        """
        super().__init__(api=api, uuid=uuid, data=data)
        self.layer_type = layer_type if isinstance(layer_type, LayerType) else LayerType(layer_type)


    def __repr__(self) -> str:
        """
        Return a string representation of the AsyncVectorLayer object.

        Returns:
            str: A string representation of the AsyncVectorLayer object.
        """
        return f"AsyncVectorLayer(uuid={self.uuid}, name={self.name}, layer_type={self.layer_type})"


    @classmethod
    async def get_vectors(cls, api: 'AsyncGeoboxClient', **kwargs) -> Union[List['AsyncVectorLayer'], int]:
        """
        [async] Get a list of vector layers with optional filtering and pagination.
        
        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.

        Keyword Args:
            include_settings (bool): Whether to include layer settings. Default is False.
            temporary (bool): Whether to return temporary layers, default is False
            q (str): Query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            search (str): Search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields (str): Comma separated list of fields for searching
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count (bool): Whether to return total count. default is False.
            skip (int): Number of layers to skip. default is 0.
            limit (int): Maximum number of layers to return. default is 10.
            user_id (int): Specific user. privileges required.
            shared (bool): Whether to return shared layers. default is False.
                
        Returns:
            List[AsyncVectorLayer] | int: A list of AsyncVectorLayer instances or the layers count if return_count is True.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layers = await AsyncVectorLayer.get_vectors(api=client, 
            ...         include_settings=True, 
            ...         skip=0, 
            ...         limit=100, 
            ...         return_count=False, 
            ...         search="my_layer",
            ...         search_fields="name, description",
            ...         order_by="name",
            ...         shared=True)
            or  
            >>>     layers = await client.get_vectors(include_settings=True, 
            ...         skip=0, 
            ...         limit=100, 
            ...         return_count=False, 
            ...         search="my_layer",
            ...         search_fields="name, description",
            ...         order_by="name",
            ...         shared=True)
        """
        params = {
            'f': 'json',
            'include_settings': kwargs.get('include_settings', False),
            'temporary': kwargs.get('temporary', False),
            'q': kwargs.get('q', None),
            'search': kwargs.get('search', None),
            'search_fields': kwargs.get('search_fields', None),
            'order_by': kwargs.get('order_by', None),
            'return_count': kwargs.get('return_count', False),
            'skip': kwargs.get('skip', 0),
            'limit': kwargs.get('limit', 10),
            'user_id': kwargs.get('user_id', None),
            'shared': kwargs.get('shared', False)
        }
        return await super()._get_list(api=api, 
                                 endpoint=cls.BASE_ENDPOINT,
                                 params=params, 
                                 factory_func=lambda api, item: AsyncVectorLayer(api, item['uuid'], LayerType(item['layer_type']), item))


    @classmethod
    async def get_vector(cls, api: 'AsyncGeoboxClient', uuid: str, user_id: int = None) -> 'AsyncVectorLayer':
        """
        [async] Get a specific vector layer by its UUID.
        
        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            uuid (str): The UUID of the layer to retrieve.
            user_id (int, optional): Specific user. privileges required.

        Returns:
            AsyncVectorLayer: The requested layer instance.
            
        Raises:
            NotFoundError: If the layer with the specified UUID is not found.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            or  
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
        """
        params = {
            'f': 'json',
            'user_id': user_id
        }
        return await super()._get_detail(api=api,
            endpoint=cls.BASE_ENDPOINT,
            uuid=uuid, 
            params=params, 
            factory_func=lambda api, item: AsyncVectorLayer(api, item['uuid'], LayerType(item['layer_type']), item))


    @classmethod
    async def get_vector_by_name(cls, api: 'AsyncGeoboxClient', name: str, user_id: int = None) -> Union['AsyncVectorLayer', None]:
        """
        [async] Get a vector layer by name

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            name (str): the name of the vector to get
            user_id (int, optional): specific user. privileges required.

        Returns:
            AsyncVectorLayer | None: returns the vector if a vector matches the given name, else None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector_by_name(client, name='test')
            or  
            >>>     layer = await client.get_vector_by_name(name='test')
        """
        layers = await cls.get_vectors(api, q=f"name = '{name}'", user_id=user_id)
        if layers and layers[0].name == name:
            return layers[0]
        else:
            return None


    @classmethod
    async def get_vectors_by_ids(cls, api: 'AsyncGeoboxClient', ids: List[int], user_id: int = None, include_settings: bool = False) -> List['AsyncVectorLayer']:
        """
        [async] Get vector layers by their IDs.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            ids (List[int]): The IDs of the layers to retrieve.
            user_id (int, optional): Specific user. privileges required.
            include_settings (bool, optional): Whether to include the layer settings. default is False.

        Returns:
            List[AsyncVectorLayer]: The list of AsyncVectorLayer instances.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layers = await AsyncVectorLayer.get_vectors_by_ids(api=client, ids=[1, 2, 3])
            or  
            >>>     layers = await client.get_vectors_by_ids(ids=[1, 2, 3])
        """
        params = {
            'ids': ids,
            'user_id': user_id,
            'include_settings': include_settings
        }
        return await super()._get_list_by_ids(api=api, 
            endpoint=f'{cls.BASE_ENDPOINT}get-layers/', 
            params=params, 
            factory_func=lambda api, item: AsyncVectorLayer(api, item['uuid'], LayerType(item['layer_type']), item))


    @classmethod
    async def create_vector(cls, 
        api: 'AsyncGeoboxClient', 
        name: str, 
        layer_type: LayerType, 
        display_name: str = None,
        description: str = None, 
        has_z: bool = False, 
        temporary: bool = False,
        fields: List = None) -> 'AsyncVectorLayer':
        """
        [async] Create a new vector layer.
        
        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            name (str): The name of the layer.
            layer_type (LayerType): The type of geometry to store.
            display_name (str, optional): A human-readable name for the layer. default is None.
            description (str, optional): A description of the layer. default is None.
            has_z (bool, optional): Whether the layer includes Z coordinates. default is False.
            temporary (bool, optional): Whether to create a temporary layer. temporary layers will be deleted after 24 hours. default is False.
            fields (List, optional): List of field definitions for the layer. default is None.
            
        Returns:
            AsyncVectorLayer: The newly created layer instance.
            
        Raises:
            ValidationError: If the layer data is invalid.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.create_vector(api=client, 
            ...         name="my_layer", 
            ...         layer_type=LayerType.Point,
            ...         display_name="My Layer",
            ...         description="This is a description of my layer",
            ...         has_z=False,
            ...         temporary=True,
            ...         fields=[{"name": "my_field", "datatype": "FieldTypeString"}])
            or  
            >>>     layer = await client.create_vector(name="my_layer", 
            ...         layer_type=LayerType.Point,
            ...         display_name="My Layer",
            ...         description="This is a description of my layer",
            ...         has_z=False,
            ...         temporary=True,
            ...         fields=[{"name": "my_field", "datatype": "FieldTypeString"}])
        """
        data = {
            "name": name,
            "layer_type": layer_type.value,
            "display_name": display_name,
            "description": description,
            "has_z": has_z,
            "temporary": temporary,
            "fields": fields
        }
        return await super()._create(api=api, 
            endpoint=cls.BASE_ENDPOINT,
            data=data, 
            factory_func=lambda api, item: AsyncVectorLayer(api, item['uuid'], layer_type, item))


    async def update(self, **kwargs) -> Dict:
        """
        [async] Update the layer's properties.
        
        Keyword Args:
            name (str): The new name for the layer.
            display_name (str): The new display name for the layer.
            description (str): The new description for the layer.
            
        Returns:
            Dict: The updated layer data.
            
        Raises:
            ValidationError: If the update data is invalid.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await layer.update(name="new_name")
            >>>     await layer.update(display_name="new_display_name")
            >>>     await layer.update(description="new_description")
        """
        data = {
            "name": kwargs.get('name'),
            "display_name": kwargs.get('display_name'),
            "description": kwargs.get('description')
        }
        return await super()._update(endpoint=self.endpoint, data=data)


    async def delete(self) -> None:
        """
        [async] Delete the layer.
                    
        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await layer.delete()
        """
        await super()._delete(endpoint=self.endpoint)


    async def make_permanent(self) -> None:
        """
        [async] Make the layer permanent.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await layer.make_permanent()
        """
        endpoint = urljoin(self.endpoint, 'makePermanent/')
        response = await self.api.post(endpoint, is_json=False)
        self._update_properties(response)


    async def share(self, users: List['AsyncUser']) -> None:
        """
        [async] Shares the layer with specified users.

        Args:
            users (List[AsyncUser]): The list of user objects to share the layer with.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     users = await client.search_users(search="John")
            >>>     await layer.share(users=users)
        """
        await super()._share(self.endpoint, users)
    

    async def unshare(self, users: List['AsyncUser']) -> None:
        """
        [async] Unshares the layer with specified users.

        Args:
            users (List[AsyncUser]): The list of user objectss to unshare the layer with.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     users = await client.search_users(search="John")
            >>>     await layer.unshare(users=users)
        """
        await super()._unshare(self.endpoint, users)


    async def get_shared_users(self, search: str = None, skip: int = 0, limit: int = 10) -> List['AsyncUser']:
        """
        [async] Retrieves the list of users the layer is shared with.

        Args:
            search (str, optional): The search query.
            skip (int, optional): The number of users to skip.
            limit (int, optional): The maximum number of users to retrieve.

        Returns:
            List[AsyncUser]: The list of shared users.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await layer.get_shared_users(search='John', skip=0, limit=10)
        """
        params = {
            'search': search,
            'skip': skip,
            'limit': limit
        }
        return await super()._get_shared_users(self.endpoint, params)
    

    async def create_version(self, name: str, display_name: str = None, description: str = None) -> 'AsyncVectorLayerVersion':
        """
        [async] Create a version from the layer

        Args:
            name (str): the name of the version.
            display_name (str, optional): the display name of the version.
            description (str, optional): the description of the version.

        Returns:
            VectorLayerVersion: the object of the version.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     version = await layer.create_version(name="my_version")
        """
        data = {
            "name": name,
            "display_name": display_name,
            "description": description
        }
        endpoint = urljoin(self.endpoint, 'versions')
        return await super()._create(self.api, endpoint, data, factory_func=lambda api, item: AsyncVectorLayerVersion(api, item['uuid'], item))


    async def get_versions(self, **kwargs) -> List['AsyncVectorLayerVersion']:
        """
        [async] Get list of versions of the current vector layer object with optional filtering and pagination.

        Keyword Args:
            q (str): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            search (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields (str): comma separated list of fields for searching.
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count (bool): Whether to return total count. default is False.
            skip (int): Number of items to skip. default is 0.
            limit (int): Number of items to return. default is 10.
            user_id (int): Specific user. privileges required.
            shared (bool): Whether to return shared versions. default is False.

        Returns:
            List[AsyncVectorLayerVersion] | int: A list of vector layer version instances or the total number of versions.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     versions = await layer.get_versions()
            or  
            >>>     versions = await layer.get_versions()
        """
        return await AsyncVectorLayerVersion.get_versions(self.api, layer_id=self.id, **kwargs)


    @property    
    def wfs(self) -> str:
        """
        Get the WFS endpoint for the layer.

        Returns:
            str: The WFS endpoint for the layer.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     layer.wfs
        """
        if self.api.access_token:
            return f'{self.api.base_url}{self.endpoint}wfs/'
        elif self.api.apikey:
            return f'{self.api.base_url}{self.endpoint}apikey:{self.api.apikey}/wfs/'


    async def get_fields(self) -> List['AsyncField']:
        """
        [async] Get all fields in the layer.
        
        Returns:
            List[AsyncField]: A list of Field instances representing the layer's fields.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     fields = await layer.get_fields()
        """
        endpoint = urljoin(self.endpoint, 'fields/')
        return await super()._get_list(api=self.api, 
            endpoint=endpoint,
            factory_func=lambda api, item: AsyncField(layer=self, data_type=FieldType(item['datatype']), field_id=item['id'], data=item))


    async def get_field(self, field_id: int) -> 'AsyncField':
        """
        [async] Get a specific field by its ID.
        
        Args:
            field_id (int, optional): The ID of the field to retrieve.

        Returns:
            Field: The requested field instance.
            
        Raises:
            NotFoundError: If the field with the specified ID is not found. 

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     field = await layer.get_field(field_id=1)
        """        
        field = next((f for f in await self.get_fields() if f.id == field_id), None)
        if not field:
            raise NotFoundError(f'Field with ID {field_id} not found in layer {self.name}')
            
        return field


    async def get_field_by_name(self, name: str) -> 'AsyncField':
        """
        [async] Get a specific field by its name.
        
        Args:
            name (str): The name of the field to retrieve.

        Returns:
            AsyncField: The requested field instance.
            
        Raises:
            NotFoundError: If the field with the specified name is not found. 

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     field = await layer.get_field_by_name(name='test')
        """        
        field = next((f for f in await self.get_fields() if f.name == name), None)
        if not field:
            raise NotFoundError(f"Field with name '{name}' not found in layer {self.name}")
            
        return field


    async def add_field(self, name: str, data_type: 'FieldType', data: Dict = {}) -> 'AsyncField':
        """
        [async] Add a new field to the layer.
        
        Args:
            name (str): The name of the new field.
            data_type (FieldType): The data type of the new field.
            data (Dict): Additional field properties (display_name, description, etc.).
            
        Returns:
            AsyncField: The newly created field instance.
            
        Raises:
            ValidationError: If the field data is invalid.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     field = await layer.add_field(name="new_field", data_type=FieldType.String)
        """
        return await AsyncField.create_field(self.api, layer=self, name=name, data_type=data_type, data=data)


    async def calculate_field(self, 
        target_field: str, 
        expression: str, 
        q: str = None, 
        bbox: List = None, 
        bbox_srid: int = None, 
        feature_ids: List = None, 
        run_async: bool = True, 
        user_id: int = None) -> Union['AsyncTask', Dict]:
        """
        [async] Calculate values for a field based on an expression.
        
        Args:
            target_field (str): The field to calculate values for.
            expression (str): The expression to use for calculation.
            q (str, optional): Query to filter features. default is None.
            bbox (List, optional): Bounding box to filter features. default is None.
            bbox_srid (int, optional): Spatial reference ID for the bounding box. default is None.
            feature_ids (List, optional): List of specific feature IDs to include. default is None
            run_async (bool, optional): Whether to run the calculation asynchronously. default is True.
            user_id (int, optional): Specific user. privileges required.
            
        Returns:
            AsyncTask | Dict: The task instance of the calculation operation or the api response if the run_async=False.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     task = await layer.calculate_field(target_field="target_field", 
            ...        expression="expression", 
            ...        q="name like 'my_layer'", 
            ...        bbox=[10, 20, 30, 40], 
            ...        bbox_srid=3857, 
            ...        feature_ids=[1, 2, 3], 
            ...        run_async=True)
        """
        data = clean_data({
            "target_field": target_field,
            "expression": expression,
            "q": q,
            "bbox": bbox,
            "bbox_srid": bbox_srid,
            "feature_ids": feature_ids,
            "run_async": run_async,
            "user_id": user_id
        })
        
        endpoint = urljoin(self.endpoint, 'calculateField/')
        response = await self.api.post(endpoint, data, is_json=False)
        if run_async:
            task = await AsyncTask.get_task(self.api, response.get('task_id'))
            return task

        return response


    async def get_features(self, geojson: bool = False, **kwargs) -> Union[List['AsyncFeature'], Dict, int]:
        """
        [async] Get features from the layer with optional filtering and pagination.
        
        Args:
            geojson (bool, optional): If True, returns the raw API response (GeoJSON dict). 
                            If False, returns a list of Feature objects. default: False.

        Keyword Args:
            quant_factor (int): Quantization factor. This parameter is only used by topojson encoder and is ignored for other formats. Higher quantizaion value means higher geometry precision. default is 1000000.
            skip (int): Number of features to skip. default is 0.
            limit (int): Maximum number of features to return. default is 100.
            user_id (int): Specific user. privileges required.
            search (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields (str): comma separated list of fields for searching.
            skip_geometry (bool): Whether to exclude geometry data. default is False.
            return_count (bool): Whether to return total count. default is False.
            feature_ids (list): Comma separated list of feature ids which should be filtered.
            select_fields (str): comma separated field names which should be included to the result. default is "[ALL]".
            skip_fields (str): comma separated field names which should be excluded from the result.
            out_srid (int): srid (epsg code) of result features. e.g. 4326. default is 3857.
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, length D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            q (str): query filter based on OGC CQL standard. e.g. Name LIKE '%GIS%' AND INTERSECTS(geometry, 'SRID=3857;POLYGON((4901948 2885079, 7049893 2885079, 7049893 4833901, 4901948 4833901, 4901948 2885079))').
            bbox (str): Bounding box to filter features by. e.g. [50.275, 35.1195, 51.4459, 36.0416].
            bbox_srid (int): srid (epsg code) of bbox. e.g. 4326. default is 3857.
                
        Returns:
            List[Feature] | Dict | int: A list of Feature instances or the geojson api response if geojson=True or the features count if return_count is True.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     features = await layer.get_features(
            ...         quant_factor=1000000, 
            ...         skip=0, 
            ...         limit=100, 
            ...         skip_geometry=False, 
            ...         return_count=False, 
            ...         select_fields="fclass, osm_id", 
            ...         out_srid=3857, 
            ...         bbox_srid=3857)
        """
        params = {
            'f': 'json',
            'quant_factor': kwargs.get('quant_factor', 1000000),
            'skip': kwargs.get('skip', 0),
            'limit': kwargs.get('limit', 100),
            'user_id': kwargs.get('user_id', None),
            'search': kwargs.get('search', None),
            'search_fields': kwargs.get('search_fields', None),
            'skip_geometry': kwargs.get('skip_geometry', False),
            'return_count': kwargs.get('return_count', False),
            'feature_ids': kwargs.get('feature_ids', None),
            'select_fields': kwargs.get('select_fields', '[ALL]'),
            'skip_fields': kwargs.get('skip_fields', None),
            'out_srid': kwargs.get('out_srid', 3857),
            'order_by': kwargs.get('order_by', None),
            'q': kwargs.get('q', None),
            'bbox': kwargs.get('bbox', None),
            'bbox_srid': kwargs.get('bbox_srid', 3857)
        }

        endpoint = f'{self.endpoint}features/'

        if geojson:
            return await self.api.get(endpoint)

        return await super()._get_list(api=self.api,
            endpoint=endpoint,
            params=params,
            factory_func=lambda api, item, srid: AsyncFeature(self, srid, item),
            geojson=True)


    async def get_feature(self, feature_id: int, out_srid: int = AsyncFeature.BASE_SRID) -> 'AsyncFeature':
        """
        [async] Get a specific feature by its ID.
        
        Args:
            feature_id (int): The ID of the feature to retrieve.
            out_srid (int, optional): Output spatial reference ID. default is 3857.

        Returns:
            Feature: The requested feature instance.
            
        Raises:
            NotFoundError: If the feature with the specified ID is not found.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     feature = await layer.get_feature(feature_id=1, out_srid=4326)
        """
        endpoint = f"{self.endpoint}features/{feature_id}"
        response = await self.api.get(endpoint)
        feature = AsyncFeature(self, data=response)
        if out_srid != AsyncFeature.BASE_SRID:
            feature.transform(out_srid)

        return feature


    async def create_feature(self, geojson: Dict, srid: int = AsyncFeature.BASE_SRID)-> 'AsyncFeature':
        """
        [async] Create a new feature in the layer.
        
        Args:
            geojson (Dict): The feature data including properties and geometry.
            srid (int, optional): the feature srid. default: 3857
            
        Returns:
            Feature: The newly created feature instance.
            
        Raises:
            ValidationError: If the feature data is invalid.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     geojson = {
            ...        "type": "Feature",
            ...        "geometry": {"type": "Point", "coordinates": [10, 20]},
            ...        "properties": {"name": "My Point"}
            ...     }
            >>>     feature = await layer.create_feature(geojson=geojson)
        """
        return await AsyncFeature.create_feature(self, geojson, srid=srid)


    async def delete_features(self, 
        q: str = None, 
        bbox: str = None, 
        bbox_srid: int = None, 
        feature_ids: List[int] = None, 
        run_async: bool = True, 
        user_id: int = None) -> Union['AsyncTask', Dict]:
        """
        [async] Delete features from the layer based on specified criteria.
        
        Args:
            q (str, optional): Query to filter features to delete.
            bbox (str, optional): Comma seprated bbox.
            bbox_srid (int, optional): Spatial reference ID for the bounding box.
            feature_ids (str, optional): Comma seprated feature ids.
            run_async (bool, optional): Whether to run the deletion asynchronously. default is True.
            user_id (int, optional): Specific user. privileges required.
            
        Returns:
            AsyncTask | Dict: The task instance of the deletion operation or the api response if run_async=False.
            
        Raises:
            ValidationError: If the deletion parameters are invalid.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await layer.delete_features(q="name like 'my_layer'", 
            ...         bbox='10, 20, 30, 40',
            ...         bbox_srid=3857, 
            ...         feature_ids='1, 2, 3', 
            ...         run_async=True)
        """
        data = clean_data({
            "q": q,
            "bbox": bbox,
            "bbox_srid": bbox_srid,
            "feature_ids": feature_ids,
            "run_async": run_async,
            "user_id": user_id
        })

        endpoint = urljoin(self.endpoint, 'deleteFeatures/')
        response = await self.api.post(endpoint, data, is_json=False)
        if run_async:
            task = await AsyncTask.get_task(self.api, response.get('task_id'))
            return task

        return response


    async def import_features(self, 
        file: 'AsyncFile', 
        input_geom_type: 'InputGeomType' = None, 
        input_layer_name: str = None, 
        input_dataset: str = None, 
        user_id: int = None, 
        input_srid: int = None, 
        file_encoding: str = "utf-8", 
        replace_domain_codes_by_values: bool = False, 
        report_errors: bool = True) -> 'AsyncTask':
        """
        [async] Import features from a file into the layer.
        
        Args:
            file (File): file object to import.
            input_geom_type (InputGeomType, optional): Type of geometry in the input file.
            input_layer_name (str, optional): Name of the layer in the input file.
            input_dataset (str, optional): Name of the dataset in the input file.
            user_id (int, optional): Specific user. privileges required.
            input_srid (int, optional): Spatial reference ID of the input data.
            file_encoding (str, optional): Character encoding of the input file.
            replace_domain_codes_by_values (bool, optional): Whether to replace domain codes with values.
            report_errors (bool, optional): Whether to report import errors.
            
        Returns:
            AsyncTask: The task instance of the import operation.
            
        Raises:
            ValidationError: If the import parameters are invalid.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     file = await client.get_file(uuid="12345678-1234-5678-1234-567812345678")
            >>>     task = await layer.import_features(file=file, 
            ...         input_geom_type=InputGeomType.POINT,
            ...         input_layer_name="my_layer",
            ...         input_dataset="my_dataset",
            ...         input_srid=3857,
            ...         file_encoding="utf-8",
            ...         replace_domain_codes_by_values=False, 
            ...         report_errors=True)
        """
        data = clean_data({
            "file_uuid": file.uuid,
            "input_layer": input_layer_name,
            "input_geom_type": input_geom_type.value if isinstance(input_geom_type, InputGeomType) else input_geom_type,
            "replace_domain_codes_by_values": replace_domain_codes_by_values,
            "input_dataset": input_dataset,
            "user_id": user_id,
            "input_srid": input_srid,
            "file_encoding": file_encoding,
            "report_errors": report_errors
        })

        endpoint = urljoin(self.endpoint, 'import/')
        response = await self.api.post(endpoint, data, is_json=False)
        task = await AsyncTask.get_task(self.api, response.get('task_id'))
        return task


    async def export_features(self, 
        out_filename: str, 
        out_format: 'FileOutputFormat', 
        replace_domain_codes_by_values: bool = False, 
        run_async: bool = True, 
        bbox: List[float] = None, 
        out_srid: int = None, 
        zipped: bool = True, 
        feature_ids: List[int] = None, 
        bbox_srid: int = None, 
        q: str = None, fields: List[str] = None) -> Union['AsyncTask', Dict]:
        """
        [async] Export features from the layer to a file.
        
        Args:
            out_filename (str): Name of the output file.
            out_format (FileOutputFormat): Format of the output file (e.g., 'Shapefile', 'GPKG', 'GeoJSON', 'CSV', 'KML', 'DXF').
            replace_domain_codes_by_values (bool, optional): Whether to replace domain codes with values.
            run_async (bool, optional): Whether to run the export asynchronously.
            bbox (List, optional): Bounding box to filter features.
            out_srid (int): Spatial reference ID for the output.
            zipped (bool, optional): Whether to compress the output file.
            feature_ids (List[int], optional): List of specific feature IDs to export.
            bbox_srid (int, optional): Spatial reference ID for the bounding box.
            q (str, optional): Query to filter features.
            fields (List[str], optional): List of fields to include in the export.
            
        Returns:
            AsyncTask | Dict: The task instance of the export operation or the api response if run_async=False.
            
        Raises:
            ValidationError: If the export parameters are invalid.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     task = await layer.export_features(out_filename="output.shp", 
            ...         out_format="shp",
            ...         replace_domain_codes_by_values=False, 
            ...         run_async=True, 
            ...         bbox=[10, 20, 30, 40], 
            ...         out_srid=3857, 
            ...         zipped=True, 
            ...         feature_ids=[1, 2, 3])
        """
        data = clean_data({
            "replace_domain_codes_by_values": replace_domain_codes_by_values,
            "out_format": out_format.value,
            "run_async": run_async,
            "bbox": bbox,
            "out_srid": out_srid,
            "zipped": zipped,
            "feature_ids": feature_ids,
            "bbox_srid": bbox_srid,
            "q": q,
            "out_filename": out_filename,
            "fields": fields
        })

        endpoint = urljoin(self.endpoint, 'export/')
        response = await self.api.post(endpoint, data, is_json=False)
        if run_async:
            task = await AsyncTask.get_task(self.api, response.get('task_id'))
            return task

        return response


    async def create_view(self, 
        name: str, 
        display_name: str = None,
        description: str = None, 
        view_filter: str = None, 
        view_extent: Dict = None, 
        view_cols: str = None) -> 'AsyncVectorLayerView':
        """
        [async] Create a view of the vector layer.

        Args:
            name (str): The name of the view.
            display_name (str, optional): The display name of the view.
            description (str, optional): The description of the view.
            view_filter (str, optional): The filter for the view.
            view_extent (List[float], optional): The extent of the view.
            view_cols (str, optional): The columns to include in the view.

        Returns:
            AsyncVectorLayerView: The created view instance.

        Raises:
            ValidationError: If the view parameters are invalid.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     view = await layer.create_view(name="my_view", 
            ...         display_name="My View", 
            ...         description="This is a view of my layer", 
            ...         view_filter="province_name = 'Tehran'", 
            ...         view_extent=[10, 20, 30, 40], 
            ...         view_cols="[ALL]")
        """
        from .view import AsyncVectorLayerView

        data = {
            "name": name,
            "display_name": display_name,
            "description": description,
            "view_filter": view_filter,
            "view_extent": view_extent,
            "view_cols": view_cols
        }

        return await super()._create(api=self.api,
            endpoint=f'{self.endpoint}views/',
            data=data,
            factory_func=lambda api, item: AsyncVectorLayerView(api, item['uuid'], self.layer_type, item))


    def get_tile_pbf_url(self, x: int = '{x}', y: int = '{y}', z: int = '{z}') -> str:
        """
        Get a vector tile pbf url for the layer.
        
        Args:
            x (int, optional): X coordinate of the tile.
            y (int, optional): Y coordinate of the tile.
            z (int, optioanl): Zoom level of the tile.

        Returns:
            str: The vector tile url.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     tile = layer.get_tile(x=10, y=20, z=1)
        """
        endpoint = f'{self.api.base_url}{self.endpoint}tiles/{z}/{x}/{y}.pbf'

        if not self.api.access_token and self.api.apikey:
            endpoint = f'{endpoint}?apikey={self.api.apikey}'

        return endpoint


    async def get_tile_json(self) -> Dict:
        """
        [async] Get the vector tile JSON configuration for the layer.
        
        Returns:
            Dict: The vector tile JSON configuration.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     tile_json = await layer.get_tile_json()
        """
        endpoint = urljoin(self.endpoint, 'tilejson.json') 
        return await self.api.get(endpoint)


    @property
    async def settings(self) -> Dict:
        """
        [async] Get the layer's settings.
        
        Returns:
            Dict: The layer settings.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     setting = await layer.setting
        """
        return await super()._get_settings(endpoint=self.endpoint)
    

    async def update_settings(self, settings: Dict) -> Dict:
        """
        [async] Update the settings

        settings (Dict): settings dictionary

        Returns:
            Dict: updated settings

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer1 = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     layer2 = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     await layer1.update_settings(await layer2.settings)
        """
        return await super()._set_settings(self.endpoint, settings)  


    async def set_settings(self, **kwargs) -> Dict:
        """
        [async] Set the settings of the Vector Layer.
        
        Keyword Args:
            title_field (str): The field to use as the title.
            domain_display_type (str): The type of domain display.
            allow_export (bool): Whether to allow export.
            editable (bool): Whether to allow editing.
            edit_geometry (bool): Whether to allow editing the geometry.
            editable_attributes (str): The attributes to allow editing.
            allow_insert (bool): Whether to allow inserting.
            allow_delete (bool): Whether to allow deleting.
            min_zoom (int): The minimum zoom level.
            max_zoom (int): The maximum zoom level.
            max_features (int): The maximum number of features.
            filter_features (bool): Whether to filter features.
            fields (List[str]): The fields to include in the layer.
            use_cache (bool): Whether to use caching.
            cache_until_zoom (int): The zoom level to cache until.
            
        Returns:
            Dict: The updated settings.
            
        Raises:
            ValidationError: If the settings data is invalid.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await layer.set_settings(title_field="name",
            ...         domain_display_type="Value",
            ...         allow_export=True,
            ...         editable=True,
            ...         edit_geometry=True,
            ...         editable_attributes="[ALL]",
            ...         allow_insert=True,
            ...         allow_delete=True,
            ...         min_zoom=0,
            ...         max_zoom=24,
            ...         max_features=65536,
            ...         filter_features=True,
            ...         fields=["id"],
            ...         use_cache=True,
            ...         cache_until_zoom=17)
        """
        general_settings = {'title_field', 'domain_display_type', 'allow_export'}
        edit_settings = {'editable', 'edit_geometry', 'editable_attributes', 'allow_insert', 'allow_delete'}
        tile_settings = {'min_zoom', 'max_zoom', 'max_features', 'filter_features', 'fields', 'use_cache', 'cache_until_zoom'}

        settings = await self.settings

        for key, value in kwargs.items():
            if key in general_settings:
                settings['general_settings'][key] = value
            elif key in edit_settings:
                settings['edit_settings'][key] = value
            elif key in tile_settings:
                settings['tile_settings'][key] = value

        return await super()._set_settings(endpoint=self.endpoint, data=settings)


    async def seed_cache(self, from_zoom: int = None, to_zoom: int = None, ignore_cache: bool = False, workers: int = 1, user_id: int = None) -> List['AsyncTask']:
        """
        [async] Seed the cache for the layer.
        
        Args:
            from_zoom (int, optional): The zoom level to start caching from.
            to_zoom (int, optional): The zoom level to stop caching at.
            ignore_cache (bool, optional): Whether to ignore the cache. default is False.
            workers (int, optional): The number of workers to use. default is 1.
            user_id (int, optional): Specific user. privileges required.

        Returns:
            List[AsyncTask]: The task instance of the cache seeding operation.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     task = await layer.seed_cache(from_zoom=0, to_zoom=10, ignore_cache=False, workers=1)
        """
        data = {
            'from_zoom': from_zoom,
            'to_zoom': to_zoom,
            'ignore_cache': ignore_cache,
            'workers': workers,
            'user_id': user_id
        }
        return await super()._seed_cache(endpoint=self.endpoint, data=data)
        

    async def clear_cache(self) -> None:
        """
        [async] Clear the layer's cache.
        
        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await layer.clear_cache()
        """
        await super()._clear_cache(endpoint=self.endpoint)
    

    @property
    async def cache_size(self) -> int:
        """
        [async] Get the size of the layer's cache.
        
        Returns:
            int: The size of the layer's cache.
            
        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await layer.cache_size
        """
        return await super()._cache_size(endpoint=self.endpoint)


    async def update_stats(self) -> None:
        """
        [async] Update the layer's statistics.
        
        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await layer.update_stats()
        """
        endpoint = urljoin(self.endpoint, 'updateStats/')
        return await self.api.post(endpoint)
    

    async def prune_edited_areas(self) -> None:
        """
        [async] Prune edited areas. This method eliminates edited areas when there are too many of them. Cache builder uses this edited areas for partial cache generating.
        
        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await layer.prune_edited_areas()
        """
        endpoint = urljoin(self.endpoint, 'prune/')
        return await self.api.post(endpoint)
    

    async def get_attachments(self, **kwargs) -> List['AsyncAttachment']:
        """
        [async] Get the resouces attachments

        Keyword Args:
            element_id (str): the id of the element with attachment.
            search (str): search term for keyword-based searching among all textual fields.
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            skip (int): Number of items to skip. default is 0.
            limit (int): Number of items to return. default is 10.
            return_count (bool): Whether to return total count. default is False.

        Returns:
            List[AsyncAttachment] | int: A list of attachments instances or the total number of attachments.

        Raises:
            TypeError: if the resource type is not supported

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await layer.get_attachments()
        """
        from .attachment import AsyncAttachment

        return await AsyncAttachment.get_attachments(self.api, resource=self, **kwargs)
    
    
    async def create_attachment(self, 
        name: str, 
        loc_x: int,
        loc_y: int,
        file: 'AsyncFile',
        feature: 'AsyncFeature' = None,
        display_name: str = None, 
        description: str = None) -> 'AsyncAttachment':
        """
        [async] Create a new Attachment.

        Args:
            name (str): The name of the scene.
            loc_x (int): x parameter of the attachment location.
            loc_y (int): y parameter of the attachment location.
            file (File): the file object.
            feature (Feature, optional): the feature object.
            display_name (str, optional): The display name of the scene.
            description (str, optional): The description of the scene.

        Returns:
            AsyncAttachment: The newly created Attachment instance.

        Raises:
            ValidationError: If the Attachment data is invalid.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await AsyncVectorLayer.get_vector(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     file = await client.get_file(uuid="12345678-1234-5678-1234-567812345678")
            >>>     await layer.create_attachment(name='test', loc_x=10, loc_y=10, file=file)
        """
        from .attachment import AsyncAttachment

        return await AsyncAttachment.create_attachment(self.api,
            name=name,
            loc_x=loc_x,
            loc_y=loc_y,
            resource=self,
            file=file,
            feature=feature,
            display_name=display_name,
            description=description)
    

    def to_sync(self, sync_client: 'GeoboxClient') -> 'VectorLayer':
        """
        Switch to sync version of the vector layer instance to have access to the sync methods

        Args:
            sync_client (GeoboxClient): The sync version of the GeoboxClient instance for making requests.

        Returns:
            VectorLayer: the sync instance of the vector layer.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.vectorlayer import AsyncVectorLayer
            >>> client = GeoboxClient()
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     layer = await AsyncVectorLayer.get_vector(async_client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     sync_layer = layer.to_sync(client)
        """
        from ..vectorlayer import VectorLayer

        return VectorLayer(api=sync_client, uuid=self.uuid, layer_type=self.layer_type, data=self.data)



from urllib.parse import urljoin
from typing import Optional, Dict, TYPE_CHECKING, Any

from .base import Base
from .utils import clean_data
from .enums import FieldType

if TYPE_CHECKING:
    from .api import GeoboxClient
    from .vectorlayer import VectorLayer
    from .aio import AsyncGeoboxClient
    from .aio.field import AsyncField


class Field(Base):

    def __init__(self, 
        layer: 'VectorLayer',
        data_type: 'FieldType',
        field_id: int = None,
        data: Optional[Dict] = {},
    ):
        """
        Constructs all the necessary attributes for the Field object.

        Args:
            layer (VectorLayer): The vector layer that the field belongs to.
            data_type (FieldType): type of the field
            field_id (int): the id of the field
            data (Dict, optional): The data of the field.
        """
        super().__init__(api=layer.api, data=data)
        self.layer = layer
        self.field_id = field_id
        if not isinstance(data_type, FieldType):
            raise ValueError("data_type must be a FieldType instance")
        self.data_type = data_type
        self.endpoint = urljoin(layer.endpoint, f'fields/{self.id}/') if self.data.get('id') else None


    def __repr__(self) -> str:
        """
        Return a string representation of the field.

        Returns:
            str: The string representation of the field.
        """
        return f"Field(id={self.id}, name={self.name}, data_type={self.data_type})"

        
    def __getattr__(self, name: str) -> Any:
        """
        Get an attribute from the resource.

        Args:
            name (str): The name of the attribute
        """
        if name == 'datatype':
            return FieldType(self.data['datatype'])
        return super().__getattr__(name)


    @property
    def domain(self) -> Dict:
        """
        Domain property

        Returns:
            Dict: domain data
        """
        return self.data.get('domain')


    @domain.setter
    def domain(self, value: Dict) -> None:
        """
        Domain property setter

        Returns:
            None
        """
        self.data['domain'] = value
        

    @classmethod
    def create_field(cls, 
        api: 'GeoboxClient', 
        layer: 'VectorLayer', 
        name: str, 
        data_type: 'FieldType', 
        data: Dict = {},
    ) -> 'Field':
        """
        Create a new field

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            layer (VectorLayer): field's layer
            name (str): name of the field
            data_type (FieldType): type of the field
            data (Dict, optional): the data of the field

        Returns:
            Field: the created field object

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.vectorlayer import VectorLayer
            >>> from geobox.field import Field
            >>> client = GeoboxClient()
            >>> layer = client.get_layer(uuid="12345678-1234-5678-1234-567812345678")
            >>> field = Field.create_field(client, layer=layer, name='test', data_type=FieldType.Integer)
        """
        data.update({
            "name": name,
            "datatype": data_type.value
        })
        endpoint = urljoin(layer.endpoint, 'fields/')
        return super()._create(api, endpoint, data, factory_func=lambda api, item: Field(layer, data_type, item['id'], item))


    def save(self) -> None:
        """
        Save the field. Creates a new field if field_id is None, updates existing field otherwise.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.field import Field
            >>> client = GeoboxClient()
            >>> layer = client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>> field = Field(layer=layer, data_type=FieldType.String)
            >>> field.save()
        """
        data = clean_data({
            "name": self.name,
            "datatype": FieldType(self.data_type).value,
            "display_name": self.data.get("display_name"),
            "description": self.data.get("description"),
            "domain": self.data.get("domain"),
            "width": self.data.get("width"),
            "hyperlink": self.data.get("hyperlink")
        })
        
        try: 
            if self.id:
                response = self.layer.api.put(self.endpoint, data)
        except AttributeError:
            endpoint = urljoin(self.layer.endpoint, 'fields/')
            response = self.layer.api.post(endpoint, data)
            self.id = response['id']
            self.endpoint = urljoin(self.layer.endpoint, f'fields/{self.id}/')
        
        self._update_properties(response)


    def delete(self) -> None:
        """
        Delete the field.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.field import Field
            >>> client = GeoboxClient()
            >>> layer = client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>> field = layer.get_field(name='test')
            >>> field.delete()
        """
        super()._delete(self.endpoint)
        self.field_id = None


    def update(self, **kwargs) -> Dict:
        """
        Update the field.

        Keyword Args:
            name (str): The name of the field.
            display_name (str): The display name of the field.
            description (str): The description of the field.
            domain (Dict): the domain of the field
            width (int): The width of the field.
            hyperlink (bool): the hyperlink field.

        Returns:
            Dict: The updated data.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.field import Field
            >>> client = GeoboxClient()
            >>> layer = client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>> field = layer.get_field(name='test')
            >>> field.update(name="my_field", display_name="My Field", description="My Field Description")
        """       
        data = {
            "name": kwargs.get('name'),
            "display_name": kwargs.get('display_name'),
            "description": kwargs.get('description'),
            "domain": kwargs.get('domain'),
            "hyperlink": kwargs.get('hyperlink')
        }
        return super()._update(self.endpoint, data)


    def get_field_unique_values(self) -> Dict:
        """
        Get the unique values of the field.

        Returns:
            Dict: The response data.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.field import Field
            >>> client = GeoboxClient()
            >>> layer = client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>> field = layer.get_field(name='test')
            >>> field.get_field_unique_values()
        """
        endpoint = urljoin(self.endpoint, 'distinct/')
        return self.layer.api.get(endpoint)


    def get_field_unique_values_numbers(self) -> int:
        """
        Get the count of unique values of the field.

        Returns:
            int: The count of the field unique values.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.field import Field
            >>> client = GeoboxClient()
            >>> layer = client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>> field = layer.get_field(name='test')
            >>> field.get_field_unique_values_numbers()
        """
        endpoint = urljoin(self.endpoint, 'distinctCount/')
        return self.layer.api.get(endpoint)


    def get_field_statistic(self, func: str) -> Dict:
        """
        Get the statistic of the field.

        Args:
            func (str): The function to apply to the field. values are: min, max, avg

        Returns:
            Dict: The response data.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.field import Field
            >>> client = GeoboxClient()
            >>> layer = client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>> field = layer.get_field(name='test')
            >>> field.get_field_statistic(func='avg')
        """
        endpoint = urljoin(self.endpoint, f'stats/?func_type={func}')
        return self.layer.api.get(endpoint)


    def update_domain(self, 
        range_domain: Dict = None, 
        list_domain: Dict = None,
    ) -> Dict:
        """
        Update field domian values

        Args:
            range_domain (Dict): a dictionary with min and max keys.
            list_domain (Dict): a dictionary containing the domain codes and values.

        Returns:
            Dict: the updated field domain

        Example:
            >>> from geobox import GeoboxClient
            >>> client = GeoboxClient()
            >>> field = client.get_vector(uuid="12345678-1234-5678-1234-567812345678").get_fields()[0]
            >>> range_d = {'min': 1, 'max': 10}
            >>> field.update_domain(range_domain = range_d)
            or
            >>> list_d = {'1': 'value1', '2': 'value2'}
            >>> field.update_domain(list_domain=list_d)
        """
        if not self.domain:
            self.domain = {'min': None, 'max': None, 'items': {}}

        if range_domain:
            self.domain['min'] = range_domain['min']
            self.domain['max'] = range_domain['max']

        if list_domain:
            self.domain['items'] = {**self.domain['items'], **list_domain}

        self.update(domain=self.domain)
        return self.domain


    def to_async(self, async_client: 'AsyncGeoboxClient') -> 'AsyncField':
        """
        Switch to async version of the field instance to have access to the async methods

        Args:
            async_client (AsyncGeoboxClient): The async version of the GeoboxClient instance for making requests.

        Returns:
            AsyncField: the async instance of the field.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.field import Field
            >>> client = GeoboxClient()
            >>> layer = client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>> field = layer.get_field(name='test')
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     async_field = field.to_async(async_client)
        """
        from .aio.field import AsyncField

        async_layer = self.layer.to_async(async_client=async_client)
        return AsyncField(layer=async_layer, data_type=self.data_type, field_id=self.field_id, data=self.data)



from urllib.parse import urljoin
from typing import Optional, Dict, List, Union, TYPE_CHECKING
import os
import mimetypes
from geobox.exception import ValidationError
import aiohttp
import sys

from .base import AsyncBase
from .task import AsyncTask
from .feature import AsyncFeature
from ..enums import FileFormat, PublishFileType, InputGeomType, FileType
from ..utils import clean_data, get_save_path, get_unique_filename
from ..exception import ApiRequestError

if TYPE_CHECKING:
    from . import AsyncGeoboxClient 
    from .user import AsyncUser
    from ..api import GeoboxClient
    from ..file import File


class AsyncFile(AsyncBase):

    BASE_ENDPOINT: str = 'files/'

    def __init__(self, 
        api: 'AsyncGeoboxClient', 
        uuid: str, 
        data: Optional[Dict] = {}):
        """
        Constructs all the necessary attributes for the File object.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance.
            uuid (str): The UUID of the file.
            data (Dict, optional): The data of the file.
        """
        super().__init__(api, uuid=uuid, data=data)


    def __repr__(self) -> str:
        """
        Return a string representation of the File object.

        Returns:
            str: A string representation of the File object.
        """
        return f"AsyncFile(uuid={self.uuid}, file_name={self.name}, file_type={self.file_type.value})"


    @property
    def layers(self) -> List[Dict]:
        """
        Get the layers of the file.

        Returns:
            List[Dict]: The layers of the file.
        
        Example:
            >>> from geobox import GeoboxClient
            >>> client = GeoboxClient()
            >>> file = File.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> file.layers
        """
        return self.data.get('layers', {}).get('layers', [])


    @property
    def file_type(self) -> 'FileType':
        """
        Get the file type

        Returns:
            FileType: the file type enumeration

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> async with AsyncGeoboxClient() as client:
            >>>     file = await File.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     file.file_type
        """
        return FileType(self.data.get('file_type'))


    @classmethod
    async def upload_file(cls, api: 'AsyncGeoboxClient', path: str, user_id: int = None, scan_archive: bool = True) -> 'AsyncFile':
        """
        [async] Upload a file to the GeoBox API.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance.
            path (str): The path to the file to upload.
            user_id (int, optional): specific user. privileges required.
            scan_archive (bool, optional): Whether to scan the archive for layers. default: True

        Returns:
            AsyncFile: The uploaded file instance.

        Raises:
            ValueError: If the file type is invalid.
            FileNotFoundError: If the file does not exist.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.file import AsyncFile
            >>> async with AsyncGeoboxClient() as client:
            >>>     file = await AsyncFile.upload_file(client, path='path/to/file.shp')
            or  
            >>>     file = await client.upload_file(path='path/to/file.shp')
        """
        if not os.path.exists(path):
            raise FileNotFoundError(f"File not found: {path}")

        FileFormat(os.path.splitext(path)[1])

        endpoint = cls.BASE_ENDPOINT


        with open(path, 'rb') as f:
            f = open(path, "rb")
            files = {'file': f}
            file_data = await api.post(endpoint, files=files, is_json=False)
        return cls(api, file_data['uuid'], file_data)
        

    @classmethod
    async def get_files(cls, api:'AsyncGeoboxClient', **kwargs) -> Union[List['AsyncFile'], int]:
        """
        [async] Retrieves a list of files.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.

        Keyword Args:
            q (str): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            search (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields (str): comma separated list of fields for searching
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D.NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count (bool): if true, the total number of results will be returned. default is False.
            skip (int): number of results to skip. default is 0.
            limit (int): number of results to return. default is 10.
            user_id (int): filter by user id.
            shared (bool): Whether to return shared files. default is False.
            
        Returns:
            List[AsyncFile] | int: A list of File objects or the total number of results.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.file import AsyncFile
            >>> async with AsyncGeoboxClient() as client:
            >>>     files = await AsyncFile.get_files(client, search_fields='name', search='GIS', order_by='name', skip=10, limit=10)
            or  
            >>>     files = await client.get_files(search_fields='name', search='GIS', order_by='name', skip=10, limit=10)
        """
        params = {
            'f': 'json',
            'q': kwargs.get('q', None),
            'search': kwargs.get('search', None),
            'search_fields': kwargs.get('search_fields', None),
            'order_by': kwargs.get('order_by', None),
            'return_count': kwargs.get('return_count', False),
            'skip': kwargs.get('skip', 0),
            'limit': kwargs.get('limit', 10),
            'user_id': kwargs.get('user_id', None),
            'shared': kwargs.get('shared', False)
        }
        return await super()._get_list(api, cls.BASE_ENDPOINT, params, factory_func=lambda api, item: cls(api, item['uuid'], item))


    @classmethod
    async def get_file(cls, api: 'AsyncGeoboxClient', uuid: str, user_id: int = None) -> 'AsyncFile':
        """
        [async] Retrieves a file by its UUID.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance.
            uuid (str): The UUID of the file.
            user_id (int, optional): specific user. privileges required.

        Returns:
            AsyncFile: The retrieved file instance.

        Raises:
            NotFoundError: If the file with the specified UUID is not found.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.file import AsyncFile
            >>> async with AsyncGeoboxClient() as client:
            >>>     file = await AsyncFile.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            or  
            >>>     file = await client.get_file(uuid="12345678-1234-5678-1234-567812345678")
        """
        params = {
            'f': 'json',
            'user_id': user_id
        }
        return await super()._get_detail(api, cls.BASE_ENDPOINT, f'{uuid}/info', params, factory_func=lambda api, item: AsyncFile(api, item['uuid'], item))


    @classmethod
    async def get_files_by_name(cls, api: 'AsyncGeoboxClient', name: str, user_id: int = None) -> List['AsyncFile']:
        """
        [async] Get files by name

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            name (str): the name of the file to get
            user_id (int, optional): specific user. privileges required.

        Returns:
            List[AsyncFile]: returns files that matches the given name

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.file import AsyncFile
            >>> async with AsyncGeoboxClient() as client:
            >>>     files = await AsyncFile.get_files_by_name(client, name='test')
            or  
            >>>     files = await client.get_files_by_name(name='test')
        """
        return await cls.get_files(api, q=f"name = '{name}'", user_id=user_id)


    def _get_file_name(self, response: aiohttp.ClientResponse) -> str:
        """
        Get the file name from the response.

        Args:
            response (aiohttp.ClientResponse): The response of the request.

        Returns:
            str: The file name 
        """
        if 'Content-Disposition' in response.headers and 'filename=' in response.headers['Content-Disposition']:
            file_name = response.headers['Content-Disposition'].split('filename=')[-1].strip().strip('"')
        else:
            content_type = response.headers.get("Content-Type", "")
            file_name = f'{self.name}.{mimetypes.guess_extension(content_type.split(";")[0])}'

        return file_name


    def _create_progress_bar(self) -> 'tqdm':
        """Creates a progress bar for the task."""
        try:
            from tqdm.auto import tqdm
        except ImportError:
            from .api import logger
            logger.warning("[tqdm] extra is required to show the progress bar. install with: pip install geobox[tqdm]")
            return None

        return tqdm(unit="B", 
            total=int(self.size), 
            file=sys.stdout,
            dynamic_ncols=True,
            desc="Downloading",
            unit_scale=True,
            unit_divisor=1024, 
            ascii=True
        )


    async def download(self, save_path: str = None, progress_bar: bool = True, file_name: str = None, overwrite: bool = False) -> str:
        """
        [async] Download a file and save it to the specified path.
        
        Args:
            save_path (str, optional): Path where the file should be saved. 
                                    If not provided, it saves to the current working directory
                                    using the original filename and appropriate extension.
            progress_bar (bool, optional): Whether to show a progress bar. default: True
            file_name (str, optional): the downloaded file name.
            overwrite (bool, optional): whether to overwrite the downloaded file if it exists on the save path. default is False.
        
        Returns:
            str: Path where the file was saved
            
        Raises:
            ValueError: If uuid is not set
            OSError: If there are issues with file operations

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.file import AsyncFile
            >>> async with AsyncGeoboxClient() as client:
            >>>     file = await AsyncFile.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await file.download(save_path='path/to/save/')
        """
        if not self.uuid:
            raise ValueError("File UUID is required to download the file")
        
        save_path = get_save_path(save_path)
        os.makedirs(save_path, exist_ok=True)

        # Use the aiohttp session directly for streaming
        endpoint = urljoin(self.api.base_url, f"{self.endpoint}download/")
        
        async with self.api.session.session.get(endpoint) as response:
            if response.status != 200:
                raise ApiRequestError(f"Failed to download file: {response.status}")
            
            file_name = self._get_file_name(response)
            full_path = os.path.join(save_path, file_name)
            if os.path.exists(full_path) and not overwrite:
                full_path = get_unique_filename(save_path, file_name)
            with open(full_path, 'wb') as f:
                pbar = self._create_progress_bar() if progress_bar else None
                
                # Use aiohttp's streaming method
                async for chunk in response.content.iter_chunked(8192):
                    f.write(chunk)
                    if pbar:
                        pbar.update(len(chunk))
                        pbar.refresh()
                
                if pbar:
                    pbar.close()

        return os.path.abspath(full_path)


    async def delete(self) -> None:
        """
        [async] Deletes the file.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.file import AsyncFile
            >>> async with AsyncGeoboxClient() as client:
            >>>     file = await AsyncFile.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await file.delete()
        """
        return await super()._delete(self.endpoint)


    async def publish(self, 
        name: str, 
        publish_as: 'PublishFileType' = None, 
        input_geom_type: 'InputGeomType' = None, 
        input_layer: str = None, 
        input_dataset: str = None, 
        user_id: int = None, 
        input_srid: int = AsyncFeature.BASE_SRID, 
        file_encoding: str = "UTF-8", 
        replace_domain_codes_by_values: bool = False, 
        report_errors: bool = True, 
        as_terrain: bool = False) -> 'AsyncTask':
        """
        [async] Publishes a file as a layer.

        Args:
            name (str): The name of the layer.
            publish_as (PublishFileType, optional): The type of layer to publish as.
            input_geom_type (InputGeomType, optional): The geometry type of the layer.
            input_layer (str, optional): The name of the input layer.
            input_dataset (str, optional): The name of the input dataset.
            user_id (int, optional): Specific user. privileges required.
            input_srid (int, optional): The SRID of the layer. default is: 3857
            file_encoding (str, optional): The encoding of the file. default is "utf-8".
            replace_domain_codes_by_values (bool, optional): Whether to replace domain codes by values. default is False.
            report_errors (bool, optional): Whether to report errors. default is True.
            as_terrain (bool, optional): Whether to publish as terrain. default is False.

        Returns:
            Task: The task object.

        Raises:
            ValueError: If the publish_as is not a valid PublishFileType.
            ValidationError: if the zipped file doesn't have any layers to publish.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.file import AsyncFile
            >>> async with AsyncGeoboxClient() as client:
            >>>     file = await AsyncFile.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await file.publish(publish_as=PublishFileType.VECTOR, 
            ...                         layer_name='my_layer', 
            ...                         input_geom_type=InputGeomType.POINT, 
            ...                         input_layer='layer1', 
            ...                         input_dataset='dataset1', 
            ...                         input_srid=4326, 
            ...                         file_encoding='UTF-8')
        """
        if not publish_as:
            # checks the file format or file first layer format to dynamically set the publish_as
            if self.file_type.value in ['GeoJSON', 'GPKG', 'DXF', 'GPX', 'Shapefile', 'KML', 'CSV', 'FileGDB'] or \
                (self.file_type.value in ['Complex'] and self.layers and \
                    FileType(self.layers[0]['format']).value in ['GeoJSON', 'GPKG', 'DXF', 'GPX', 'Shapefile', 'KML', 'CSV', 'FileGDB']):
                publish_as = PublishFileType.VECTOR

            elif self.file_type.value in ['GeoTIFF'] or \
                    (self.file_type.value in ['Complex'] and self.layers and \
                        FileType(self.layers[0]['format']).value in ['GeoTIFF']):
                publish_as = PublishFileType.RASTER

            elif self.file_type.value in ['GLB'] or \
                    (self.file_type.value in ['Complex'] and self.layers and \
                        FileType(self.layers[0]['format']).value in ['GLB']):
                publish_as = PublishFileType.MODEL3D

            elif self.file_type.value in ['ThreedTiles']:
                publish_as = PublishFileType.Tiles3D

            else:
                raise ValidationError('Unknown format')

        data = clean_data({
            "publish_as": publish_as.value if isinstance(publish_as, PublishFileType) else publish_as,
            "layer_name": name,
            "input_layer": self.layers[0]['layer'] if not input_layer and self.layers else input_layer,
            "input_geom_type": input_geom_type.value if isinstance(input_geom_type, InputGeomType) else input_geom_type,
            "replace_domain_codes_by_values": replace_domain_codes_by_values,
            "input_dataset": self.layers[0]['dataset'] if not input_layer and self.layers else input_dataset,
            "user_id": user_id,
            "input_srid": input_srid,
            "file_encoding": file_encoding,
            "report_errors": report_errors,
            "as_terrain": as_terrain
        })
        endpoint = urljoin(self.endpoint, 'publish/')
        response = await self.api.post(endpoint, data, is_json=False)
        task = await AsyncTask.get_task(self.api, response.get('task_id'))
        return task


    async def share(self, users: List['AsyncUser']) -> None:
        """
        [async] Shares the file with specified users.

        Args:
            users (List[User]): The list of users objects to share the file with.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.file import AsyncFile
            >>> async with AsyncGeoboxClient() as client:
            >>>     file = await AsyncFile.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     users = await client.search_users(search='John')
            >>>     await file.share(users=users)
        """
        await super()._share(self.endpoint, users)
    

    async def unshare(self, users: List['AsyncUser']) -> None:
        """
        [async] Unshares the file with specified users.

        Args:
            users (List[AsyncUser]): The list of users objects to unshare the file with.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.file import AsyncFile
            >>> async with AsyncGeoboxClient() as client:
            >>>     file = await AsyncFile.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await users = client.search_users(search='John')
            >>>     await file.unshare(users=users)
        """
        await super()._unshare(self.endpoint, users)


    async def get_shared_users(self, search: str = None, skip: int = 0, limit: int = 10) -> List['AsyncUser']:
        """
        [async] Retrieves the list of users the file is shared with.

        Args:
            search (str, optional): The search query.
            skip (int, optional): The number of users to skip.
            limit (int, optional): The maximum number of users to retrieve.

        Returns:
            List[AsyncUser]: The list of shared users.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.file import AsyncFile
            >>> async with AsyncGeoboxClient() as client:
            >>>     file = await AsyncFile.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await file.get_shared_users(search='John', skip=0, limit=10)
        """
        params = {
            'search': search,
            'skip': skip,
            'limit': limit
        }
        return await super()._get_shared_users(self.endpoint, params)


    def to_sync(self, sync_client: 'GeoboxClient') -> 'File':
        """
        Switch to sync version of the file instance to have access to the sync methods

        Args:
            sync_client (GeoboxClient): The sync version of the GeoboxClient instance for making requests.

        Returns:
            File: the async instance of the file.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.file import AsyncFile
            >>> client = GeoboxClient()
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     file = await AsyncFile.get_file(async_client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     sync_file = file.to_sync(client)
        """
        from ..file import File

        return File(api=sync_client, uuid=self.uuid, data=self.data)